package pt.uc.dei.crf_ma;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Random;
import java.util.TreeMap;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pt.uc.dei.crf_ma.evaluator.EvaluatorResults;
import pt.uc.dei.crf_ma.evaluator.MultiSegmentationEvaluator;
import pt.uc.dei.crf_ma.evaluator.PerClassAccuracyEvaluator;
import pt.uc.dei.crf_ma.evaluator.TermExtractorEvaluator;
import pt.uc.dei.crf_ma.evaluator.WordEvaluator;
import pt.uc.dei.crf_ma.pipe.ArticlesAsInstancesFileIterator;
import pt.uc.dei.crf_ma.pipe.DbpediaLexiconMembership;
import pt.uc.dei.crf_ma.pipe.MultiTokenLexiconMembership;
import pt.uc.dei.crf_ma.pipe.OffsetConjunctions;
import pt.uc.dei.crf_ma.pipe.POSTagger;
import pt.uc.dei.crf_ma.pipe.RegexMatches;
import pt.uc.dei.crf_ma.pipe.SentencesAsInstancesFileIterator;
import pt.uc.dei.crf_ma.pipe.SequencePrintingPipe;
import pt.uc.dei.crf_ma.pipe.SimpleTaggerSentence2TokenSequence;
import pt.uc.dei.crf_ma.pipe.SingleTokenLexiconMembership;
import pt.uc.dei.crf_ma.types.TermExtractorInstance;
import cc.mallet.fst.CRFCacheStaleIndicator;
import cc.mallet.fst.CRFOptimizableByBatchLabelLikelihood;
import cc.mallet.fst.CRFTrainerByValueGradients;
import cc.mallet.fst.HMM;
import cc.mallet.fst.HMMTrainerByLikelihood;
import cc.mallet.fst.SumLatticeDefault;
import cc.mallet.fst.ThreadedOptimizable;
import cc.mallet.fst.TransducerEvaluator;
import cc.mallet.optimize.Optimizable;
import cc.mallet.pipe.Pipe;
import cc.mallet.pipe.SerialPipes;
import cc.mallet.pipe.TokenSequence2FeatureSequence;
import cc.mallet.pipe.TokenSequence2FeatureVectorSequence;
import cc.mallet.pipe.TokenSequenceLowercase;
import cc.mallet.pipe.tsf.FeaturesInWindow;
import cc.mallet.pipe.tsf.TokenFirstPosition;
import cc.mallet.pipe.tsf.TokenText;
import cc.mallet.types.FeatureVectorSequence;
import cc.mallet.types.InstanceList;
import cc.mallet.types.Sequence;
import cc.mallet.types.InstanceList.CrossValidationIterator;

public class SequenceLabeller {

	public static final int DEFAULT_NUM_THREADS = 4;
	public static final double DEFAULT_GUASSIAN_VARIANCE = 10.0;
	public static final double DEFAULT_TRAINING_FRACTION = 0.9;
	public static final int DEFAULT_RANDOM_SEED = 0;

	public static final int DEFAULT_EVAL_PERIOD = 100;
	public static final int DEFAULT_NFOLDS = 0;
	public static final String DEFAULT_LABEL = "O";
	public static final String[] START_TAGS = {"I-KEYWORD",};
	public static final String[] CONTINUE_TAGS = {"C-KEYWORD",};
	public static final String FORBIDDEN_PATTERN = "O,C-*";
	public static final int DEFAULT_ITERATIONS = Integer.MAX_VALUE;
	public static final String TOKEN_FNAME = "WORD=";
	public static final String[] REAL_VALUED_FEATURES = {
		SentencesAsInstancesFileIterator.GLOBAL_SENTENCE_NUMBER_FEATURE,
		SentencesAsInstancesFileIterator.PARAGRAPH_NUMBER_FEATURE,
		SentencesAsInstancesFileIterator.RELATIVE_SENTENCE_NUMBER_FEATURE,
		SentencesAsInstancesFileIterator.TFIDF_FEATURE,
		SentencesAsInstancesFileIterator.IDF_FEATURE,
		SentencesAsInstancesFileIterator.TF_FEATURE,
	};  
	private int [][] conjunctions;

	private Logger logger = LoggerFactory.getLogger(SequenceLabeller.class);

	private CRF crfmodel;
	private SerialPipes featurePipeline;

	private double trainingFraction;
	private double gaussianVariance;
	private int randomSeed;
	private int iterations;
	private int evalPeriod;
	private int nfolds;
	private boolean useArticlesAsInstances = false;
	private boolean useTFIDF = true;

	// debug pipes
	private boolean debugPipes = false;
	private PrintWriter debugPipesOutput = null;
	private boolean useHMM = false;
	
	public int usedIterations;

	public SequenceLabeller() {

		this.trainingFraction = DEFAULT_TRAINING_FRACTION;
		this.gaussianVariance = DEFAULT_GUASSIAN_VARIANCE;
		this.randomSeed = DEFAULT_RANDOM_SEED;
		this.iterations = DEFAULT_ITERATIONS;
		this.evalPeriod = DEFAULT_EVAL_PERIOD;
		this.nfolds = DEFAULT_NFOLDS;

		this.initilizePipes();
	}

	public SequenceLabeller(CRF crf) {
		this.crfmodel = crf; 
		this.featurePipeline = (SerialPipes) crfmodel.getInputPipe();
	}

	/*
	 * @param filename Model filename
	 * 
	 * Read the TermExtractor model from a file
	 * 
	 * @return A TermExtractor model
	 */
	public static SequenceLabeller loadModel(String filename) throws Exception {
		SequenceLabeller result;

		ObjectInputStream s = new ObjectInputStream(
				new GZIPInputStream(new FileInputStream(filename)));

		CRF crf = (CRF) s.readObject();
		s.close();

		result = new SequenceLabeller(crf);
		return result;
	}

	public HashMap<String, Integer> getDatasetSizes(HashMap<String, String> trainDataFiles) throws IOException {
		featurePipeline.setTargetProcessing(true);
		HashMap<String, Integer> datasetSizes = new HashMap<String, Integer>();
		for (String id : trainDataFiles.keySet()) {
			InstanceList tmp = loadInstancesFromDirectory(trainDataFiles.get(id), ".*?\\.txt", "\\[", "\\]");
			datasetSizes.put(id, tmp.size());
		}
		return datasetSizes;
	}

	public void trainModel(String trainDataFilename) throws IOException {
		trainModel(trainDataFilename, DEFAULT_NUM_THREADS);
	}

	public void trainModel(HashMap<String, String> trainDataFiles, int nthreads, HashMap<String, Double> trainDataPercentages) throws IOException {
		trainModelForClustering(trainDataFiles, nthreads, Integer.MAX_VALUE, trainDataPercentages);
	}

	public void trainModel(HashMap<String, String> trainDataFiles, int nthreads) throws IOException {
		HashMap<String, Double> trainDataPercentages = new HashMap<String, Double>();
		for (String id : trainDataFiles.keySet()) {
			trainDataPercentages.put(id, 1.0);
		}
		trainModelForClustering(trainDataFiles, nthreads, Integer.MAX_VALUE, trainDataPercentages);
	}

	public TreeMap<String, InstanceList> loadInstancesFromFolder(String inputFolder, TreeMap<String, InstanceList> distinctInstances) {
		TreeMap<String, InstanceList> taggerInstances = new TreeMap<String, InstanceList>();
		if (!inputFolder.endsWith("/")) {
			inputFolder += "/";
		}
		File dir = new File(inputFolder);
		String[] children = dir.list();
		if (children == null) {
			throw new RuntimeException("Data directory not found: " + inputFolder);
		} else {
			for (int i = 0; i < children.length; i++) {
				String filename = children[i];
				InstanceList instances = loadInstancesFromDirectory(inputFolder + "" + filename, ".*?\\.txt", "\\[", "\\]");
				if (distinctInstances != null) {
					// keep information about all the distinct instances found on the distinctInstances HashMap
					if (instances.size() > 0) {
						InstanceList instancesCopy = (InstanceList) instances.cloneEmpty();
						for (int j=0; j<instances.size(); j++) {
							TermExtractorInstance instance = (TermExtractorInstance) instances.get(j);
							TermExtractorInstance instanceCopy = new TermExtractorInstance(instance.getData(), instance.getTarget(), instance.getName(), instance.getSource());
							instanceCopy.setInstanceId(instance.getInstanceId());
							instanceCopy.setTokens(instance.getTokens());
							instanceCopy.setLabels(null);
							instancesCopy.add(instanceCopy);
						}
						String instanceId = ((TermExtractorInstance) instances.get(0)).getInstanceId();
						if (!distinctInstances.containsKey(instanceId)) {
							distinctInstances.put(instanceId, instancesCopy);
						} else {
							// check instancelist sizes for consistency 
							if (distinctInstances.get(instanceId).size() != instances.size()) {
								//throw new RuntimeException("Different instancelist sizes for instanceId:" + instanceId + " -> "+distinctInstances.get(instanceId).size()+" != "+tmp.size());
							}
						}
					}
				}
				taggerInstances.put(filename, instances);
			}
		}
		return taggerInstances;
	}

	public void trainModelForClustering(HashMap<String, String> trainDataFiles, int nthreads, int numIterations, HashMap<String, Double> trainDataPercentages) throws IOException {
		// The training file contains tags (aka targets)
		featurePipeline.setTargetProcessing(true);
		// read all instances from trainning file
		//InstanceList instances = loadInstances(trainDataFilename);
		InstanceList instances = new InstanceList(this.featurePipeline);
		for (String id : trainDataFiles.keySet()) {
			double percentageToUse = trainDataPercentages.get(id);
			System.out.println("[CLUSTERING] 		Loading instances: " + trainDataFiles.get(id) + " percentage=" + percentageToUse);
			InstanceList tmp = loadInstancesFromDirectory(trainDataFiles.get(id), ".*?\\.txt", "\\[", "\\]");
			double[] proprotions = {percentageToUse, 1.0-percentageToUse};
			InstanceList sample = tmp.split(new Random(0), proprotions)[0];
			System.out.println("[CLUSTERING] 		Original size: " + tmp.size() + " sample size=" + sample.size());
			instances.addAll(sample);
			System.out.println("[CLUSTERING] 		New train dataset size: " + instances.size());
		}
	}

	public InstanceList trainModel(TreeMap<String, TreeMap<String, InstanceList>> trainingInstances, int nthreads, int numIterations, TreeMap<String, Double> trainDataPercentages) throws IOException {
		InstanceList allInstances = new InstanceList(this.featurePipeline);
		int instanceIndex = 0;
		for (String taggerName : trainingInstances.keySet()) {
			//System.out.println("[MultipleTaggers] ADDING TAGGER:" + taggerName);
			TreeMap<String, InstanceList> taggerInstancesMap = trainingInstances.get(taggerName);
			InstanceList taggerInstances = new InstanceList(this.featurePipeline);
			for (String filename : taggerInstancesMap.keySet()) {
				if (filename.equals(".DS_Store")) {
					continue;
				}
				taggerInstances.addAll(taggerInstancesMap.get(filename));
				//System.out.println("[MultipleTaggers]    TAGGER FILENAME:" + filename);
			}
			
			// old version
//			if (trainDataPercentages == null) {
//				allInstances.addAll(taggerInstances);
//			} else {
//				double percentageToUse = trainDataPercentages.get(taggerName);
//				while (percentageToUse > 1) {
//					percentageToUse--;
//					allInstances.addAll(taggerInstances);
//				}
//				double[] proprotions = {percentageToUse, 1.0-percentageToUse};
//				InstanceList sample = taggerInstances.split(new Random(0), proprotions)[0];
//				allInstances.addAll(sample);
//			}
			
			// new version
			allInstances.addAll(taggerInstances);
			for (; instanceIndex < allInstances.size(); instanceIndex++) {
				allInstances.setInstanceWeight(instanceIndex, trainDataPercentages.get(taggerName));
			}
		}
		//allInstances.shuffle(new Random(1)); // shuffle destroys instance weights
		//System.out.println("[MultipleTaggers] training data size: " + allInstances.size());
		
		// The training file contains tags (aka targets)
		featurePipeline.setTargetProcessing(true);

		/* 
		 * split instances into training and testing data
		 * to prevent overfitting
		 */

		if (useHMM) {
			logger.info("Using HMMs");
		} else {
			logger.info("Using CRFs");
		}

		if (nfolds == 0) {
			trainModelPercentageSplit(allInstances, nthreads, numIterations);
		} else {
			trainModelCrossValidation(allInstances, nthreads, numIterations);
		}

		//        PrintWriter pw = new PrintWriter("crfweights.txt");
		//        crfmodel.print(pw);
		//        pw.close();
		return allInstances;
	}

	public void trainModel(String trainDataFilename, int nthreads) throws IOException {
		trainModel(trainDataFilename, nthreads, Integer.MAX_VALUE);
	}

	public void trainModel(String trainDataFilename, int nthreads, int numIterations) throws IOException {
		// The training file contains tags (aka targets)
		featurePipeline.setTargetProcessing(true);
		// read all instances from trainning file
		//InstanceList instances = loadInstances(trainDataFilename);
		InstanceList instances = loadInstancesFromDirectory(trainDataFilename, ".*?\\.txt", "\\[", "\\]");

		/* 
		 * split instances into training and testing data
		 * to prevent overfitting
		 */

		if (useHMM) {
			logger.info("Using HMMs");
		} else {
			logger.info("Using CRFs");
		}

		if (nfolds == 0) {
			trainModelPercentageSplit(instances, nthreads, numIterations);
		} else {
			trainModelCrossValidation(instances, nthreads, numIterations);
		}

		PrintWriter pw = new PrintWriter("crfweights.txt");
		crfmodel.print(pw);
		pw.close();
	}

	public void trainModelPercentageSplit(InstanceList instances, int nthreads, int numIterations) throws IOException {
		/* 
		 * split instances into training and testing data
		 * to prevent overfitting
		 */
		//System.out.println("[MultipleTaggers] Before percentage split: " + instances.getInstanceWeight(0));
		InstanceList trainData, testData;
		if (trainingFraction >= 1.0) {
			trainData = instances;
			testData = new InstanceList(featurePipeline);
		} else {
			Random r = new Random(randomSeed);
			InstanceList[] trainingLists =
				instances.split(r, new double[]{trainingFraction,
						1 - trainingFraction});
			trainData = trainingLists[0];
			testData = trainingLists[1];
		}
		//System.out.println("[MultipleTaggers] After percentage split: " + trainData.getInstanceWeight(0));

		logger.info("Spliting {} instances into {} ({}) training instances and {} ({}) testing instances.", 
				new Object[] {instances.size(), trainData.size(), trainingFraction, testData.size(), 1.0f-trainingFraction});

		logger.info("Random Seed for split: {}", randomSeed);
		logger.info("Number of features in training data: {}", featurePipeline.getDataAlphabet().size());
		logger.info("Initiating training with {} threads.", nthreads);
		logger.info("Gaussian variance: {}", gaussianVariance);
		logger.info("Training model until convergence or after reaching {} iterations", iterations);
		logger.info("Evaluating period in every {} iterations.", evalPeriod);

		if (useHMM) {
			trainHMMModel(trainData, testData, 0, numIterations);
		} else {
			trainCRFModel(trainData, testData, nthreads, 0, numIterations);
		}
	}

	public void trainModelCrossValidation(InstanceList instances, int nthreads, int numIterations) throws IOException {
		logger.info("Doing cross validation with {} folds.", nfolds);

		// train classifiers using cross validation
		CrossValidationIterator cvIter = instances.crossValidationIterator(nfolds, randomSeed);
		int fold = 0;
		ArrayList<EvaluatorResults[]> allResults = new ArrayList<EvaluatorResults[]>();
		while (cvIter.hasNext()) {
			fold++;
			InstanceList[] foldData = cvIter.nextSplit();
			logger.info("Fold: " + fold + " / " + nfolds + " (trainData size=" + foldData[0].size() + ", testData size=" + foldData[1].size() + ")");
			EvaluatorResults[] results;
			if (useHMM) {
				results  = trainHMMModel(foldData[0], foldData[1], fold, numIterations);
			} else {
				results  = trainCRFModel(foldData[0], foldData[1], nthreads, fold, numIterations);
			}
			allResults.add(results);
		}

		// compute average results over all folds
		ArrayList<EvaluatorResults> avgResults = new ArrayList<EvaluatorResults>();
		for (int i = 0; i < allResults.get(0).length; i++) {
			avgResults.add(new EvaluatorResults(0, 0, 0, 0, 0, 0, 0, 0, 0));
		}
		PrintWriter pw = new PrintWriter("results.txt");
		if (useArticlesAsInstances)
			pw.println("INSTANCES = ARTICLES");
		else
			pw.println("INSTANCES = SENTENCES");
		pw.println("FEATURE PIPELINE: " + pipelineToString() + "\n");
		//for (EvaluatorResults[] results : allResults) {
		for (fold = 0; fold < allResults.size(); fold++) {
			EvaluatorResults[] results = allResults.get(fold);
			for (int i=0; i<avgResults.size(); i++) {
				avgResults.get(i).setDescription(results[i].getDescription());
				results[i].setDescription("FOLD: " + (fold + 1) + " " + results[i].getDescription());
				logger.info(results[i].toString());
				pw.println(results[i].toString());
				avgResults.get(i).setAccuracy(avgResults.get(i).getAccuracy() + results[i].getAccuracy());
				avgResults.get(i).setF1(avgResults.get(i).getF1() + results[i].getF1());
				avgResults.get(i).setNumCorrectSegments(avgResults.get(i).getNumCorrectSegments() + results[i].getNumCorrectSegments());
				avgResults.get(i).setNumCorrectTokens(avgResults.get(i).getNumCorrectTokens() + results[i].getNumCorrectTokens());
				avgResults.get(i).setNumPredictedSegments(avgResults.get(i).getNumPredictedSegments() + results[i].getNumPredictedSegments());
				avgResults.get(i).setNumTrueSegments(avgResults.get(i).getNumTrueSegments() + results[i].getNumTrueSegments());
				avgResults.get(i).setPrecision(avgResults.get(i).getPrecision() + results[i].getPrecision());
				avgResults.get(i).setRecall(avgResults.get(i).getRecall() + results[i].getRecall());
				avgResults.get(i).setTotalTokens(avgResults.get(i).getTotalTokens() + results[i].getTotalTokens());
				avgResults.get(i).setNumIterations(avgResults.get(i).getNumIterations() + results[i].getNumIterations());
			}
		}

		for (int i=0; i<avgResults.size(); i++) {
			avgResults.get(i).setAccuracy(avgResults.get(i).getAccuracy() / nfolds);
			avgResults.get(i).setF1(avgResults.get(i).getF1() / nfolds);
			avgResults.get(i).setNumCorrectSegments(avgResults.get(i).getNumCorrectSegments() / nfolds);
			avgResults.get(i).setNumCorrectTokens(avgResults.get(i).getNumCorrectTokens() / nfolds);
			avgResults.get(i).setNumPredictedSegments(avgResults.get(i).getNumPredictedSegments() / nfolds);
			avgResults.get(i).setNumTrueSegments(avgResults.get(i).getNumTrueSegments() / nfolds);
			avgResults.get(i).setPrecision(avgResults.get(i).getPrecision() / nfolds);
			avgResults.get(i).setRecall(avgResults.get(i).getRecall()  / nfolds);
			avgResults.get(i).setTotalTokens(avgResults.get(i).getTotalTokens() / nfolds);
			avgResults.get(i).setNumIterations(avgResults.get(i).getNumIterations() / nfolds);
			avgResults.get(i).setDescription("FINAL RESULTS (AVG): " + avgResults.get(i).getDescription());

			pw.println("\n" + avgResults.get(i).toString());
			logger.info(avgResults.get(i).toString());
		}
		for (int i=0; i<avgResults.size(); i++) {
			System.out.print(avgResults.get(i).getPrecision() + "\t" + avgResults.get(i).getRecall() + "\t" + avgResults.get(i).getF1() + "\t");
		}

		pw.flush();
		pw.close();
	}

	public EvaluatorResults[] trainHMMModel(InstanceList trainData, InstanceList testData, int fold, int numIterations) {
		HMM hmm = new HMM(featurePipeline, null);
		hmm.addStatesForLabelsConnectedAsIn(trainData);
		//hmm.addStatesForBiLabelsConnectedAsIn(trainData);

		HMMTrainerByLikelihood trainer = new HMMTrainerByLikelihood(hmm);

		TransducerEvaluator [] evals;
		evals = new TransducerEvaluator [] { 
				new MultiSegmentationEvaluator(new InstanceList[]{trainData, testData}, new String[]{"Training", "Testing"},
						START_TAGS, CONTINUE_TAGS),
						new PerClassAccuracyEvaluator(new InstanceList[]{trainData, testData}, new String[]{"Training", "Testing"}),
						//        		new TermExtractorEvaluator(new InstanceList[]{trainData, testData}, new String[]{"Training", "Testing"},
						//        		START_TAGS, CONTINUE_TAGS),
						//        		new WordEvaluator(new InstanceList[]{trainData, testData}, new String[]{"Training", "Testing"},
						//        		START_TAGS, CONTINUE_TAGS),
		};

		boolean converged;
		iterations = 2000; // HMMs training does not seem to converge...
		EvaluatorResults[] evaluatorResults = new EvaluatorResults[2];
		int i;
		for (i = 1; i <= numIterations; i++) {
			converged = trainer.train(trainData, 1);
			if ((evalPeriod != 0 && i % evalPeriod == 0)) {
				logger.info("Evaluating on iteration {} for fold {}:", i, fold);
				for (int j=0;j<evals.length;j++)
					evals[j].evaluate(trainer);
			} else if (converged) {
				break;
			}
		}

		logger.info("Evaluating on iteration {} for fold {}:", i, fold);
		for (int j=0;j<evals.length;j++) {
			evals[j].evaluate(trainer);
			if (evals[j] instanceof MultiSegmentationEvaluator) {
				evaluatorResults[0] = ((MultiSegmentationEvaluator) evals[j]).evaluateInstanceListWithResults(trainer, trainData, "FINAL EVALUATION", true);
				evaluatorResults[0].setNumIterations(i);
				evaluatorResults[1] = ((MultiSegmentationEvaluator) evals[j]).evaluateInstanceListWithResults(trainer, testData, "FINAL EVALUATION", true);
				evaluatorResults[1].setNumIterations(i);
			} 
		}

		return evaluatorResults;
	}

	public EvaluatorResults[] trainCRFModel(InstanceList trainData, InstanceList testData, int nthreads, int fold, int numIterations) {
		/*
		 *  initialize mallet CRF
		 */
		crfmodel = new CRF(featurePipeline, null);

		// construct the finite state machine
		crfmodel.addStatesForLabelsConnectedAsIn(trainData);
		//crfmodel.addStatesForBiLabelsConnectedAsIn(trainData);
		// initialize model's weights
		crfmodel.setWeightsDimensionAsIn(trainData, false);

		/* here is where the actual training happens */
		//TODO: need to check this parameters of initial weights
		//CRFTrainerByLabelLikelihood trainer = new CRFTrainerByLabelLikelihood(crfmodel);
		//CRFTrainerByThreadedLabelLikelihood trainer = new CRFTrainerByThreadedLabelLikelihood(crfmodel, nthreads);

		// CRF trainer
		// old version - parallel version
//		CRFOptimizableByBatchLabelLikelihood batchOptLabel =
//			new CRFOptimizableByBatchLabelLikelihood(crfmodel, trainData, nthreads);
//		ThreadedOptimizable optLabel = new ThreadedOptimizable(
//				batchOptLabel, trainData, crfmodel.getParameters().getNumFactors(),
//				new CRFCacheStaleIndicator(crfmodel));
//		Optimizable.ByGradientValue[] opts =
//			new Optimizable.ByGradientValue[]{optLabel};
//		// by default, use L-BFGS as the optimizer
//		CRFTrainerByValueGradients trainer =
//			new CRFTrainerByValueGradients(crfmodel, opts);
//		batchOptLabel.setGaussianPriorVariance(gaussianVariance);
		
		// new version - with weights
		CRFTrainerByLabelLikelihood trainer = new CRFTrainerByLabelLikelihood(crfmodel);
		trainer.setGaussianPriorVariance(gaussianVariance);
		
		/* 
		 * specify which evaluators to use
		 */
		TransducerEvaluator [] evals;
		evals = new TransducerEvaluator [] { 
				new MultiSegmentationEvaluator(new InstanceList[]{trainData, testData}, new String[]{"Training", "Testing"}, START_TAGS, CONTINUE_TAGS),
				new PerClassAccuracyEvaluator(new InstanceList[]{trainData, testData}, new String[]{"Training", "Testing"}),
				new TermExtractorEvaluator(new InstanceList[]{trainData, testData}, new String[]{"Training", "Testing"}, START_TAGS, CONTINUE_TAGS),
				new WordEvaluator(new InstanceList[]{trainData, testData}, new String[]{"Training", "Testing"}, START_TAGS, CONTINUE_TAGS),
		};

		boolean converged;
		EvaluatorResults[] evaluatorResults = new EvaluatorResults[4];
		int iteration;
		for (iteration = 1; iteration <= numIterations; iteration+=5) {
			System.out.println("[MultipleTaggers]- CRF iteration=" + iteration);
			converged = trainer.train(trainData, 5);
			if ((evalPeriod != 0 && iteration % evalPeriod == 0)) {
				logger.info("Evaluating on iteration {} for fold {}:", iteration, fold);
				for (int j=0;j<evals.length;j++)
					evals[j].evaluate(trainer);
			} else if (converged) {
				logger.info("Evaluating on iteration {} for fold {}:", iteration, fold);
				break;
			}
		}
		for (int j=0;j<evals.length;j++) {
			evals[j].evaluate(trainer);
			if (evals[j] instanceof MultiSegmentationEvaluator) {
				evaluatorResults[0] = ((MultiSegmentationEvaluator) evals[j]).evaluateInstanceListWithResults(trainer, trainData, "DESC", true);
				evaluatorResults[0].setNumIterations(iteration);
				evaluatorResults[0].setDescription("MultiSegmentationEvaluator: Training Data");
				evaluatorResults[1] = ((MultiSegmentationEvaluator) evals[j]).evaluateInstanceListWithResults(trainer, testData, "DESC", true);
				evaluatorResults[1].setNumIterations(iteration);
				evaluatorResults[1].setDescription("MultiSegmentationEvaluator: Testing Data");
			} else if (evals[j] instanceof TermExtractorEvaluator) {
				evaluatorResults[2] = ((TermExtractorEvaluator) evals[j]).evaluateInstanceListWithResults(trainer, testData, "DESC", true);
				evaluatorResults[2].setNumIterations(iteration);
				evaluatorResults[2].setDescription("TermExtractorEvaluator: Testing Data");
			} else if (evals[j] instanceof WordEvaluator) {
				evaluatorResults[3] = ((WordEvaluator) evals[j]).evaluateInstanceListWithResults(trainer, testData, "DESC", true);
				evaluatorResults[3].setNumIterations(iteration);
				evaluatorResults[3].setDescription("WordEvaluator: Testing Data");
			}
		}
		usedIterations = iteration;
		//optLabel.shutdown();

		return evaluatorResults;
	}

	public TreeMap<String, InstanceList> applyModelToFile(String testDataFilename, String outputFilename, boolean prettyFormat) throws IOException {
		logger.info("Starting test phase!");
		PrintWriter outputFile = null;
		if(outputFilename == null) 
			outputFile = new PrintWriter(System.out);
		else 
			outputFile = new PrintWriter(outputFilename);

		featurePipeline.setTargetProcessing(false);
		InstanceList testData = this.loadInstances(testDataFilename);
		TreeMap<String, InstanceList> testDataMap = new TreeMap<String, InstanceList>();
		testDataMap.put(testDataFilename, testData);
		return applyModelToInstances(testDataMap, outputFile, prettyFormat);
	}

	public TreeMap<String, InstanceList> applyModelToInstances(TreeMap<String, InstanceList> testDataMap, PrintWriter outputFile, boolean prettyFormat) {
		TreeMap<String, InstanceList> testDataPredMap = new TreeMap<String, InstanceList>();
		for (String filename : testDataMap.keySet()) {
			if (filename.equals(".DS_Store")) {
				continue;
			}
			//System.out.println("APPLYING TO FILE:" + filename);
			InstanceList testData = testDataMap.get(filename);
			InstanceList testDataPred = testData.cloneEmpty();
			for (int i = 0; i < testData.size(); i++) {
				TermExtractorInstance testInstance = (TermExtractorInstance) testData.get(i);
				ArrayList<String> inputTokens = testInstance.getTokens();
				ArrayList<String> labels = new ArrayList<String>();
				FeatureVectorSequence input = (FeatureVectorSequence) testInstance.getData();
				Sequence output = crfmodel.transduce(input);

				// Prediction probabilities
				double logScore = new SumLatticeDefault(crfmodel, input, output).getTotalWeight();
				double logZ = new SumLatticeDefault(crfmodel, input).getTotalWeight();
				double prob = Math.exp(logScore - logZ);

				boolean error = false;
				if (output.size() != input.size())
					error = true;
				//System.out.println("[MultipleTaggers] output.size(): " + output.size());
				//System.out.println("[MultipleTaggers] input.size(): " + input.size());
				//System.out.println("[MultipleTaggers] inputTokens.length: " + inputTokens.size());

				if (!error) {
					StringBuffer buf = new StringBuffer("Sequence Prob: " + prob + "\n");
					for (int j = 0; j < inputTokens.size(); j++) {
						String token = inputTokens.get(j);
						//System.out.println("[MultipleTaggers] DEBUG: " + token + " - " + output.get(j).toString());
						String label = output.get(j).toString();
						labels.add(label);
						if(prettyFormat) {
							String nextLabel = null;
							if(j+1 < output.size()) 
								nextLabel = output.get(j+1).toString();

							if (label.equals(SequenceLabeller.DEFAULT_LABEL)) {
								buf.append(token);
							} else if (label.equals(SequenceLabeller.START_TAGS[0])) {
								buf.append("[").append(token);
							} else if (label.equals(SequenceLabeller.CONTINUE_TAGS[0])) {
								buf.append(token);
							}

							if((label.equals(SequenceLabeller.START_TAGS[0]) || label.equals(SequenceLabeller.CONTINUE_TAGS[0])) 
									&& (nextLabel == null || nextLabel.equals(SequenceLabeller.DEFAULT_LABEL))) {
								buf.append("]");
							}

							buf.append(" ");
						} else {
							buf.append(token).append("\t").append(label).append("\n");
						}
					}
					if (outputFile != null) {
						outputFile.println(buf.append("\n").toString());
					}
				}
				
				TermExtractorInstance testInstanceCopy = new TermExtractorInstance(testInstance.getData(), testInstance.getTarget(), testInstance.getName(), testInstance.getSource());
				testInstanceCopy.setLabels(labels);
				testInstanceCopy.setTokens(testInstance.getTokens());
				testDataPred.add(testInstanceCopy);
			}
			testDataPredMap.put(filename, testDataPred);
		}

		if (outputFile != null) {
			outputFile.flush();
			outputFile.close();
		}

		return testDataPredMap;
	}

	public double[] labelProbabilities(String testDataFilename) throws IOException {
		logger.info("Starting test phase!");

		featurePipeline.setTargetProcessing(true);
		InstanceList testData = loadInstancesFromDirectory(testDataFilename, ".*?\\.txt", "\\[", "\\]");
		double logScoreSum = 0;
		double logZSum = 0;
		double probSum = 0;
		for (int i = 0; i < testData.size(); i++) {
			FeatureVectorSequence input = (FeatureVectorSequence) testData.get(i).getData();
			Sequence output = crfmodel.transduce(input);

			// Prediction probabilities
			double logScore = new SumLatticeDefault(crfmodel, input, output).getTotalWeight();
			double logZ = new SumLatticeDefault(crfmodel, input).getTotalWeight();
			double prob = Math.exp(logScore - logZ);
			logScoreSum += logScore;
			logZSum += logZ;
			probSum += prob;
		}

		double size = (double) testData.size();
		return new double[]{logScoreSum/size, logZSum/size, probSum/size};
	}

	public double loglikelihood(String dataFilename) throws IOException {
		featurePipeline.setTargetProcessing(true);
		InstanceList instances = loadInstancesFromDirectory(dataFilename, ".*?\\.txt", "\\[", "\\]");
		return loglikelihood(instances);
	}
	
	public double loglikelihood(InstanceList instances) throws IOException {
		CRFOptimizableByLabelLikelihood optimizable = new CRFOptimizableByLabelLikelihood(crfmodel, instances);
		//double dataLogLikelihood = optimizable.getValue();
		double dataLogLikelihood = optimizable.getExpectationValue();

		//System.out.println("[CLUSTERING] data size=" + testData.size());
		//System.out.println("[CLUSTERING] data loglikelihood=" + dataLogLikelihood);
		//return dataLogLikelihood;
		
		return dataLogLikelihood / instances.size();
	}

	public void saveModel(String modelFileName) throws Exception {
		// Implement own 
		if(crfmodel == null) 
			throw new RuntimeException("Model not trained. Nothing to save");

		// gzip model
		ObjectOutputStream oos = new ObjectOutputStream(
				new GZIPOutputStream(new FileOutputStream(modelFileName)));
		oos.writeObject(crfmodel);
		oos.close();
	}

	/* 
	 * Private methods 
	 */

	private void initilizePipes() {
		String CAPS = "[\\p{Lu}]";
		String LOW = "[\\p{Ll}]";
		String ALPHA = "[\\p{Lu}\\p{Ll}]";
		String ALPHANUM = "[\\p{Lu}\\p{Ll}\\p{Nd}]";
		String PUNT = "[,\\.;:?!()]";
		String QUOTE = "[\"`']";

		// offset conjunctions
		conjunctions = new int[10][];
		conjunctions[0] = new int[]{-1};
		conjunctions[1] = new int[]{1};
		conjunctions[2] = new int[]{-2, -1};
		conjunctions[3] = new int[]{1, 2};
		conjunctions[4] = new int[]{-2};
		conjunctions[5] = new int[]{2};
		conjunctions[6] = new int[]{-3};
		conjunctions[7] = new int[]{3};
		conjunctions[8] = new int[]{-3, -2, -1};
		conjunctions[9] = new int[]{1, 2, 3};

		if (useHMM) {
			featurePipeline = new SerialPipes(new Pipe [] {
					new SimpleTaggerSentence2TokenSequence(true, REAL_VALUED_FEATURES),
					new TokenSequence2FeatureSequence()
			});
		} else {
			featurePipeline = new SerialPipes(new Pipe [] {

					//Sentence2TokenSequence pipe
					new SimpleTaggerSentence2TokenSequence(true, REAL_VALUED_FEATURES),
					//	        new AnnotatedText2TokenSequence(false, "\\[", "\\]"),

					new TokenFirstPosition("FIRSTTOKEN"),

					new POSTagger("POS="),

					// Caps features
					new RegexMatches("CAPITALIZED", Pattern.compile(CAPS + LOW + "*")),
					new RegexMatches("INITCAP", Pattern.compile(CAPS + ".*")),
					new RegexMatches("ALLCAPS", Pattern.compile(CAPS + "+")),
					new RegexMatches("MIXEDCAPS", Pattern.compile("[A-Z][a-z]+[A-Z][A-Za-z]*")),

					// Number features
					new RegexMatches("CONTAINSDIGITS", Pattern.compile(".*[0-9].*")),
					new RegexMatches("ALLDIGITS", Pattern.compile("[0-9]+")),
					new RegexMatches("NUMERICAL", Pattern.compile("[-0-9]+[\\.,]+[0-9\\.,]+")),
					new RegexMatches("STARTSNUMBER", Pattern.compile("^[0-9].*")),

					// Morphologic features
					new RegexMatches("MULTIDOTS", Pattern.compile("\\.\\.+")),
					new RegexMatches("HYPHENATED", Pattern.compile(".*\\-.*")),
					new RegexMatches("DOLLARSIGN", Pattern.compile(".*\\$.*")),
					new RegexMatches("ENDSINDOT", Pattern.compile("[^\\.]+.*?\\.")),
					new RegexMatches("CONTAINSDASH", Pattern.compile(ALPHANUM + "+-" + ALPHANUM + "*")),
					new RegexMatches("ACRO", Pattern.compile("[A-Z][A-Z\\.]*\\.[A-Z\\.]*")),
					new RegexMatches("LONELYINITIAL", Pattern.compile(CAPS + "\\.")),
					new RegexMatches("SINGLECHAR", Pattern.compile(ALPHA)),
					new RegexMatches("CAPLETTER", Pattern.compile("[A-Z]")),
					new RegexMatches("PUNC", Pattern.compile(PUNT)),
					new RegexMatches("ENDPUNC", Pattern.compile("[?!\\.]")),
					new RegexMatches("QUOTE", Pattern.compile(QUOTE)),

					// My features
					new RegexMatches("APOSTROFE", Pattern.compile("[^']'.*")),
					new RegexMatches("YEAR", Pattern.compile("[12][0-9][0-9][0-9]")),
					new RegexMatches("EMAIL", Pattern.compile("[a-zA-Z0-9\\._%+-]+@[a-zA-Z0-9\\.-]+(\\.[a-zA-Z]{2,4})?")),

					// lexicon memberships
					//		        new SingleTokenLexiconMembership("LEXICON=FirstNames", "/pipes/LexiconMembership/First_Names.txt.gz", false),
					//		        new SingleTokenLexiconMembership("LEXICON=Surnames", "/pipes/LexiconMembership/Surnames.txt.gz", false),
					//		        
					//		        new MultiTokenLexiconMembership("LEXICON=Albums", "/pipes/LexiconMembership/dbpedia/albums.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Albums", "/pipes/LexiconMembership/dbpedia/albums_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Artists", "/pipes/LexiconMembership/dbpedia/artists.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Artists", "/pipes/LexiconMembership/dbpedia/artists_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Books", "/pipes/LexiconMembership/dbpedia/books.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Books", "/pipes/LexiconMembership/dbpedia/books_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Cities", "/pipes/LexiconMembership/dbpedia/cities.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Cities", "/pipes/LexiconMembership/dbpedia/cities_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Companies", "/pipes/LexiconMembership/dbpedia/companies.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Companies", "/pipes/LexiconMembership/dbpedia/companies_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Countries", "/pipes/LexiconMembership/dbpedia/countries.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Countries", "/pipes/LexiconMembership/dbpedia/countries_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Magazines", "/pipes/LexiconMembership/dbpedia/magazines.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Magazines", "/pipes/LexiconMembership/dbpedia/magazines_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Musicals", "/pipes/LexiconMembership/dbpedia/musicals.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Musicals", "/pipes/LexiconMembership/dbpedia/musicals_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Newspapers", "/pipes/LexiconMembership/dbpedia/newpapers.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Newspapers", "/pipes/LexiconMembership/dbpedia/newpapers_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Occupations", "/pipes/LexiconMembership/dbpedia/occupations.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Organisations", "/pipes/LexiconMembership/dbpedia/organisations.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Organisations", "/pipes/LexiconMembership/dbpedia/organisations_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Persons", "/pipes/LexiconMembership/dbpedia/persons.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Persons", "/pipes/LexiconMembership/dbpedia/persons_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Places", "/pipes/LexiconMembership/dbpedia/places.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Places", "/pipes/LexiconMembership/dbpedia/places_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Singles", "/pipes/LexiconMembership/dbpedia/singles.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Singles", "/pipes/LexiconMembership/dbpedia/singles_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Softwares", "/pipes/LexiconMembership/dbpedia/softwares.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Softwares", "/pipes/LexiconMembership/dbpedia/softwares_title.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Songs", "/pipes/LexiconMembership/dbpedia/songs.txt.gz", false),
					//		        new MultiTokenLexiconMembership("LEXICON=Songs", "/pipes/LexiconMembership/dbpedia/songs_title.txt.gz", false),

					// lexicons with pagerank
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Albums)", "/pipes/LexiconMembership/dbpedia_pagerank/albums_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Artists)", "/pipes/LexiconMembership/dbpedia_pagerank/artists_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Books)", "/pipes/LexiconMembership/dbpedia_pagerank/books_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Cities)", "/pipes/LexiconMembership/dbpedia_pagerank/cities_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Companies)", "/pipes/LexiconMembership/dbpedia_pagerank/companies_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Countries)", "/pipes/LexiconMembership/dbpedia_pagerank/countries_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Magazines)", "/pipes/LexiconMembership/dbpedia_pagerank/magazines_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Musicals)", "/pipes/LexiconMembership/dbpedia_pagerank/musicals_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Newspapers)", "/pipes/LexiconMembership/dbpedia_pagerank/newpapers_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Occupations)", "/pipes/LexiconMembership/dbpedia_pagerank/occupations_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Organisations)", "/pipes/LexiconMembership/dbpedia_pagerank/organisations_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Persons)", "/pipes/LexiconMembership/dbpedia_pagerank/persons_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Places)", "/pipes/LexiconMembership/dbpedia_pagerank/places_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Places)", "/pipes/LexiconMembership/dbpedia_pagerank/places_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Singles)", "/pipes/LexiconMembership/dbpedia_pagerank/singles_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Softwares)", "/pipes/LexiconMembership/dbpedia_pagerank/softwares_title_pagerank.txt.gz", false),
					//	        new DbpediaLexiconMembership("DBPEDIALEXICON(Songs)", "/pipes/LexiconMembership/dbpedia_pagerank/songs_title_pagerank.txt.gz", false),

					//new DbpediaLexiconMembership("PAGERANK", "/pipes/LexiconMembership/wiki-pageranks_with_titles_normalized.txt.gz", false),

					new TokenSequenceLowercase(),

					/* Make the word a feature. */
					new TokenText("WORD="),

					new OffsetConjunctions(true, Pattern.compile("POS=.*"),
							conjunctions),
							new OffsetConjunctions(true, Pattern.compile("WORD=.*"),
									conjunctions),

									/* FeatureInWindow features. */
									new FeaturesInWindow("WINDOW1=",-1,1, Pattern.compile("WORD=.*"),true),
									new FeaturesInWindow("WINDOW2=",-1,1, Pattern.compile("POS=.*"),true),
									new FeaturesInWindow("WINDOW3=",-1,1, Pattern.compile("[A-Za-z]+=?[A-Za-z]*"),true),
									new FeaturesInWindow("WINDOW4=",-1,1, Pattern.compile("WORD=.*|POS=.*"),true),

									//  	        new OffsetConjunctions(true,Pattern.compile("WORD=.*|POS=.*"),
									//  	              conjunctions),

									new TokenSequence2FeatureVectorSequence()
			});
		}

		// write to the log the pipes that are used
		logger.info(pipelineToString());
	}

	private InstanceList loadInstances(String filename) throws IOException {
		throw new UnsupportedOperationException("NOT IMPLEMENTED YET");
	}

	private InstanceList loadInstancesFromDirectory(String directoryPath, String validFilesRegex, String startTagPattern, String endTagPattern) {
		if (useHMM) {
			debugPipes = false; // HMMs don't allow the "normal" debugging of pipes
		}
		if (debugPipes) {
			try {
				debugPipesOutput = new PrintWriter("pipes.txt");
				featurePipeline.pipes().add((new SequencePrintingPipe(debugPipesOutput, REAL_VALUED_FEATURES)));
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			}
		}

		InstanceList instances = new InstanceList(this.featurePipeline);
		if (useArticlesAsInstances) {
			//logger.info("Instances = Articles");
			instances.addThruPipe(new ArticlesAsInstancesFileIterator(directoryPath, validFilesRegex, startTagPattern, endTagPattern, useTFIDF));
		} else {
			//logger.info("Instances = Sentences");
			//instances.addThruPipe(new SentencesAsInstancesFileIterator(directoryPath, validFilesRegex, startTagPattern, endTagPattern, useTFIDF));
			instances.addThruPipe(new SentencesAsInstancesFileIterator(directoryPath, validFilesRegex, startTagPattern, endTagPattern, useTFIDF));
		}

		if (debugPipes && debugPipesOutput != null) {
			debugPipesOutput.flush();
			debugPipesOutput.close();
			featurePipeline.pipes().remove(featurePipeline.pipes().size()-1);
		}

		return instances;
	}

	private String pipelineToString() {
		// write to the log the pipes that are used
		StringBuffer buf = new StringBuffer("Feature Pipeline:\n");
		buf.append("Conjunctions: ");
		for (int i = 0; i<conjunctions.length; i++) {
			buf.append("{");
			for (int j=0; j<conjunctions[i].length; j++) {
				buf.append(conjunctions[i][j]);
				if (j != conjunctions[i].length - 1)
					buf.append(",");
			}
			buf.append("}");
			if (i != conjunctions.length - 1)
				buf.append(",");
		}
		buf.append("\n");

		// LinkedHashMap keeps the insertion order
		LinkedHashMap<String, ArrayList<String>> pipes = new LinkedHashMap<String, ArrayList<String>>();
		if (useTFIDF) {
			pipes.put("TFIDF", null);
		}
		for (int i=0;i<featurePipeline.size();i++) {
			Pipe p = featurePipeline.getPipe(i);
			String featureName = p.getClass().getSimpleName();
			if (p instanceof SingleTokenLexiconMembership) {
				String featureOpts = ((SingleTokenLexiconMembership) p).getName().replace("LEXICON=", "");
				if (pipes.containsKey(featureName)) {
					pipes.get(featureName).add(featureOpts);
				} else {
					pipes.put(featureName, new ArrayList<String>(Arrays.asList(featureOpts)));
				}
			} else if (p instanceof MultiTokenLexiconMembership) {
				String featureOpts = ((MultiTokenLexiconMembership) p).getName().replace("LEXICON=", "");
				if (pipes.containsKey(featureName)) {
					pipes.get(featureName).add(featureOpts);
				} else {
					pipes.put(featureName, new ArrayList<String>(Arrays.asList(featureOpts)));
				}
			} else if (p instanceof DbpediaLexiconMembership) {
				String featureOpts = ((DbpediaLexiconMembership) p).getName();
				if (pipes.containsKey(featureName)) {
					pipes.get(featureName).add(featureOpts);
				} else {
					pipes.put(featureName, new ArrayList<String>(Arrays.asList(featureOpts)));
				}
			} else if (p instanceof RegexMatches) {
				String featureOpts = ((RegexMatches) p).getFeature();
				if (pipes.containsKey(featureName)) {
					pipes.get(featureName).add(featureOpts);
				} else {
					pipes.put(featureName, new ArrayList<String>(Arrays.asList(featureOpts)));
				}
			} else if (p instanceof OffsetConjunctions) {
				String featureOpts = ((OffsetConjunctions) p).getFeatureRegex().pattern();
				if (pipes.containsKey(featureName)) {
					pipes.get(featureName).add(featureOpts);
				} else {
					pipes.put(featureName, new ArrayList<String>(Arrays.asList(featureOpts)));
				}
			} else if (p instanceof TokenSequence2FeatureVectorSequence || p instanceof SimpleTaggerSentence2TokenSequence) {
				// ignore these
			} else {
				pipes.put(featureName, null);
			}
		}

		for (String featureName : pipes.keySet()) {
			ArrayList<String> opts = pipes.get(featureName);
			if (opts != null) {
				buf.append(featureName + "(");
				for (int i=0; i<opts.size(); i++) {
					buf.append(opts.get(i));
					if (i != opts.size() - 1)
						buf.append(", ");
				}
				buf.append(") ");
			} else {
				buf.append(featureName + " ");
			}
		}
		return buf.toString();
	}

	/* 
	 * Getters and Setters
	 * 
	 */

	public double getTrainingFraction() {
		return trainingFraction;
	}

	public void setTrainingFraction(double trainingFraction) {
		this.trainingFraction = trainingFraction;
	}

	public double getGaussianVariance() {
		return gaussianVariance;
	}

	public void setGaussianVariance(double gaussianVariance) {
		this.gaussianVariance = gaussianVariance;
	}

	public int getRandomSeed() {
		return randomSeed;
	}

	public void setRandomSeed(int randomSeed) {
		this.randomSeed = randomSeed;
	}

	public int getIterations() {
		return iterations;
	}

	public void setIterations(int iterations) {
		this.iterations = iterations;
	}

	public int getEvalPeriod() {
		return evalPeriod;
	}

	public void setEvalPeriod(int evalPeriod) {
		this.evalPeriod = evalPeriod;
	}

	public int getNFolds() {
		return nfolds;
	}

	public void setNFolds(int n) {
		this.nfolds = n;
	}

	public boolean isDebugPipes() {
		return debugPipes;
	}

	public void setDebugPipes(boolean debugPipes) {
		this.debugPipes = debugPipes;
	}

	public SerialPipes getFeaturePipeline() {
		return featurePipeline;
	}

	public void setFeaturePipeline(SerialPipes featurePipeline) {
		this.featurePipeline = featurePipeline;
	}

	

}
