package pt.uc.dei.crf_ma.clustering;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;

import cc.mallet.types.MatrixOps;
import pt.uc.dei.crf_ma.SequenceLabeller;

public class CRFFuzzyKMeansClustering {

	private static final long serialVersionUID = 8526472295622776148L;

	private Random randomGenerator;
	private int numClusters;
	private int maxIterations;
	private HashMap<String, String> dataFiles;
	private HashMap<String, Integer> dataSizes;
	private int numInstances;
	private int numThreads = 4;
	private HashMap<String, ArrayList<Assignment>> assignments;
	private KCluster[] clusters;

	public CRFFuzzyKMeansClustering(int numClusters, int maxIterations,
			String inputFolder, int numThreads) {
		this.numClusters = numClusters;
		this.maxIterations = maxIterations;
		this.dataFiles = getDataFiles(inputFolder);
		this.numThreads = numThreads;
		this.numInstances = dataFiles.size();
		this.dataSizes = getDataSizes(this.dataFiles);
	}

	private HashMap<String, String> getDataFiles(String inputFolder) {
		HashMap<String, String> dataFiles = new HashMap<String, String>();

		File dir = new File(inputFolder);
		if (!inputFolder.endsWith("/")) {
			inputFolder += "/";
		}
		String[] children = dir.list();
		if (children == null) {
			throw new RuntimeException("Data directory not found: "
					+ inputFolder);
		} else {
			for (int i = 0; i < children.length; i++) {
				// Get filename of file or directory
				String filename = children[i];
				if (!filename.equals(".DS_Store")
						&& !filename.equals("._.DS_Store")) {
					dataFiles.put(filename, new String(inputFolder + filename));
				}
			}
		}

		return dataFiles;
	}

	private HashMap<String, Integer> getDataSizes(HashMap<String, String> dataFiles) {
		System.out.println("[CLUSTERING] Determining dataset sizes...");
		SequenceLabeller te = new SequenceLabeller();
		try {
			this.dataSizes = te.getDatasetSizes(dataFiles);
		} catch (IOException e) {
			e.printStackTrace();
		}
		return this.dataSizes;
	}

	public KCluster[] randomInitializationFull() {
		this.randomGenerator = new Random(1);
		System.out.println("[CLUSTERING] Num instances: " + this.numInstances);
		System.out.println("[CLUSTERING] Num clusters: " + this.numClusters);
		System.out.println("[CLUSTERING] Max iterations: " + this.maxIterations);
		System.out.println("[CLUSTERING] Initializing clusters with random data (full initialization)...");
		this.clusters = new KCluster[this.numClusters];
		for (int i = 0; i < this.numClusters; i++) {
			this.clusters[i] = new KCluster();
			this.clusters[i].setClusterId(i);
			SequenceLabeller te = new SequenceLabeller();

			// set the special options to the TermExtractor
			te.setEvalPeriod(100);
			te.setNFolds(0); // no cross-validation
			te.setIterations(20);
			te.setTrainingFraction(1.0);
			te.setRandomSeed(0);
			// te.setDebugPipes(true);

			this.clusters[i].setCrf(te);
		}
		this.assignments = new HashMap<String, ArrayList<Assignment>>();
		for (String id : this.dataFiles.keySet()) {
			String dataFile = this.dataFiles.get(id);
			int rand = this.randomGenerator.nextInt(this.numClusters);
			this.assignments.put(id, new ArrayList<Assignment>());
			for (KCluster cluster : this.clusters) {
				if (cluster.getClusterId() == rand) {
					this.assignments.get(id).add(new Assignment(id, cluster, 1.0));
				} else {
					this.assignments.get(id).add(new Assignment(id, cluster, 0.0));
				}
			}
			this.clusters[rand].addInstance(id, dataFile);
		}

		for (String id : this.assignments.keySet()) {
			for (Assignment assignment : assignments.get(id)) {
				System.out.println("[CLUSTERING] 	user="+id + " -> cluster="+assignment.getCluster().getClusterId() + " membership="+assignment.getMembership());
			}
		}

		return this.clusters;
	}

	public KCluster[] randomInitializationPartial() {
		this.randomGenerator = new Random(0);
		System.out.println("[CLUSTERING] Num instances: " + this.numInstances);
		System.out.println("[CLUSTERING] Num clusters: " + this.numClusters);
		System.out
				.println("[CLUSTERING] Max iterations: " + this.maxIterations);
		System.out
				.println("[CLUSTERING] Initializing clusters with random data (partial initialization)...");
		this.clusters = new KCluster[this.numClusters];
		ArrayList<Integer> pickedNumbers = new ArrayList<Integer>();
		ArrayList<String> ids = new ArrayList<String>(this.dataFiles.keySet());
		for (int i = 0; i < this.numClusters; i++) {
			this.clusters[i] = new KCluster();

			int rand = this.randomGenerator.nextInt(this.numInstances);
			while (pickedNumbers.contains(rand)) {
				rand = this.randomGenerator.nextInt(this.numInstances);
			}
			String id = ids.get(rand);
			this.clusters[i].addInstance(id, this.dataFiles.get(id));

			SequenceLabeller te = new SequenceLabeller();
			te.setEvalPeriod(100);
			te.setNFolds(0); // no cross-validation
			te.setTrainingFraction(1.0);
			te.setRandomSeed(0);
			// te.setDebugPipes(true);
			this.clusters[i].setCrf(te);
		}

		for (KCluster cluster : this.clusters) {
			System.out.print("[CLUSTERING] 	Initilizing cluster "
					+ cluster.getClusterId() + " with the instances: ");
			for (String id : cluster.getInstances().keySet()) {
				System.out.print(id + ", ");
			}
			System.out.println();
		}

		return this.clusters;
	}

	public KCluster[] randomInitializationSmart() {
		this.randomGenerator = new Random(0);
		System.out.println("[CLUSTERING] Num instances: " + this.numInstances);
		System.out.println("[CLUSTERING] Num clusters: " + this.numClusters);
		System.out.println("[CLUSTERING] Max iterations: " + this.maxIterations);
		System.out.println("[CLUSTERING] Initializing clusters with random data (smart initialization)...");
		this.clusters = new KCluster[this.numClusters];
		ArrayList<Integer> pickedNumbers = new ArrayList<Integer>();
		ArrayList<String> ids = new ArrayList<String>(this.dataFiles.keySet());
		double[][] loglikelihoods = new double[this.numClusters][this.numInstances];
		double[][] normalizedLoglikelihoods = new double[this.numClusters][this.numInstances];

		// Randomly assign first instance
		// Initialize cluster info
		int nextDataPoint = this.randomGenerator.nextInt(this.numInstances);
		for (int k = 0; k < this.numClusters; k++) {
			this.clusters[k] = new KCluster();
			pickedNumbers.add(nextDataPoint);
			String instance = ids.get(nextDataPoint);
			System.out.println("[CLUSTERING] Using instance=" + instance
					+ " to initialize cluster=" + k);
			this.clusters[k]
					.addInstance(instance, this.dataFiles.get(instance));
			SequenceLabeller te = new SequenceLabeller();
			te.setEvalPeriod(100);
			te.setNFolds(0); // no cross-validation
			te.setTrainingFraction(1.0);
			te.setRandomSeed(0);
			// te.setDebugPipes(true);
			this.clusters[k].setCrf(te);

			if (k == this.numClusters - 1) {
				continue; // skip next steps for the last cluster
			}

			// Train model
			te.setIterations(Integer.MAX_VALUE); // variable number of
													// iterations
			te.setGaussianVariance(1.0); // fixed variance
			try {
				te.trainModel(this.clusters[k].getInstances(), numThreads);
			} catch (IOException e) {
				e.printStackTrace();
			} catch (Exception e) {
				e.printStackTrace();
			}
			this.clusters[k].setCrf(te);

			// Calculate loglikelihoods
			for (int i = 0; i < this.numInstances; i++) {
				String inst = ids.get(i);
				String dataFile = this.dataFiles.get(inst);
				try {
					double ll = this.clusters[k].getCrf().loglikelihood(
							dataFile);
					System.out.println("[CLUSTERING] 	instance=" + inst
							+ " cluster=" + this.clusters[k].getClusterId()
							+ " ll=" + ll);
					loglikelihoods[k][i] = ll;
				} catch (IOException e) {
					e.printStackTrace();
					System.exit(1);
				}
			}

			// // Normalize loglikelihoods to the [0,1] interval
			// double min = MatrixOps.min(loglikelihoods[k]);
			// double max = MatrixOps.max(loglikelihoods[k]);
			// for (int i=0; i<loglikelihoods[k].length; i++) {
			// normalizedLoglikelihoods[k][i] = (loglikelihoods[k][i] - max) /
			// (max - min);
			// System.out.println("[CLUSTERING] 	normalized=" +
			// normalizedLoglikelihoods[k][i]);
			// }

			// Determine furthest data point
			double minLoglikelihoodSum = Double.MAX_VALUE;
			int furthestInstance = 0;
			for (int i = 0; i < loglikelihoods[k].length; i++) {
				double sum = 0;
				for (int j = 0; j <= k; j++) {
					sum += loglikelihoods[j][i];
				}
				sum /= (k + 1);
				System.out.println("[CLUSTERING] 	instance=" + ids.get(i)
						+ " cluster=" + this.clusters[k].getClusterId()
						+ " sum=" + sum);
				if (sum < minLoglikelihoodSum && !pickedNumbers.contains(i)) {
					minLoglikelihoodSum = sum;
					furthestInstance = i;
				}
			}
			nextDataPoint = furthestInstance;
		}

		for (KCluster cluster : this.clusters) {
			System.out.print("[CLUSTERING] 	Initilizing cluster "
					+ cluster.getClusterId() + " with the instances: ");
			for (String id : cluster.getInstances().keySet()) {
				System.out.print(id + ", ");
			}
			System.out.println();
		}

		return this.clusters;
	}

	public void clusterData() {
		//this.clusters = randomInitializationPartial();
		this.clusters = randomInitializationFull();
		//this.clusters = randomInitializationSmart();

		PrintWriter pwAssignments = null;
		try {
			pwAssignments = new PrintWriter("assignments.txt");
			pwAssignments.println("cluster_assignments = []");
			pwAssignments.println("assignments0 = {}");
			for (String id : this.assignments.keySet()) {
				Assignment bestAssignment = null;
				for (Assignment assignment : this.assignments.get(id)) {
					if (bestAssignment == null || assignment.getMembership() > bestAssignment.getMembership()) {
						bestAssignment = assignment;
					}
				}
				pwAssignments.println("assignments0['" + id + "'] = "
						+ bestAssignment.getCluster().getClusterId());
			}
			pwAssignments.println("cluster_assignments.append(assignments0)");
			pwAssignments.flush();
		} catch (FileNotFoundException e1) {
			e1.printStackTrace();
		}

		System.out.println("[CLUSTERING] Clustering started...");
		double currentGlobalLikelihoodFuzzy = 0;
		double previousGlobalLikelihoodFuzzy = 0;
		double currentGlobalLikelihoodHard = 0;
		double previousGlobalLikelihoodHard = 0;
		double[] previousGlobalLikelihoodsHard = new double[this.maxIterations];
		boolean converged = false;
		for (int iteration = 1; iteration <= this.maxIterations && !converged; iteration++) {
			System.out.println("[CLUSTERING] Iteration: " + iteration);
			
			currentGlobalLikelihoodFuzzy = 0;
			currentGlobalLikelihoodHard = 0;

			// Get previous cluster assignments
			HashMap<String, ArrayList<Assignment>> previousAssignments = new HashMap<String, ArrayList<Assignment>>();
			for (String id : this.assignments.keySet()) {
				previousAssignments.put(id, assignments.get(id));
			}

			double maxTrainingDataSize = Integer.MIN_VALUE;
			HashMap[][] trainingData = new HashMap[this.numClusters][2];
			double[] trainingDataSizes = new double[this.numClusters];
			for (int i=0; i<numClusters; i++) {
				// Prepare training data
				KCluster cluster = this.clusters[i];
				double trainingDataSize = 0;
				HashMap<String, String> trainingInstances = new HashMap<String, String>();
				HashMap<String, Double> trainingPercentages = new HashMap<String, Double>();
				for (String id : this.assignments.keySet()) {
					String instance = this.dataFiles.get(id);
					Assignment assignment = null;
					for (Assignment a : this.assignments.get(id)) {
						if (a.getCluster().getClusterId() == cluster.getClusterId()) {
							assignment = a;
						}
					}
					if (assignment.getMembership() > 0) {
						System.out.println("[CLUSTERING] 	Adding instance " + id + " for train with membership=" + assignment.getMembership());
						trainingInstances.put(id, instance);
						trainingPercentages.put(id, assignment.getMembership());
						double dataSize = this.dataSizes.get(id) * assignment.getMembership();
						trainingDataSize += dataSize;
					}
				}
				trainingData[i][0] = trainingInstances;
				trainingData[i][1] = trainingPercentages;
				trainingDataSizes[i] = trainingDataSize;
				if (trainingDataSize > maxTrainingDataSize) {
					maxTrainingDataSize = trainingDataSize;
				}
			}

			// Train clusters
			int trainingIterations = 500;
			//int trainingIterations = (int) Math.pow(2, iteration-1);
			System.out.println("[CLUSTERING] 	Trying with training iterations=" + trainingIterations);
			for (int i=0; i<numClusters; i++) {
				KCluster cluster = this.clusters[i];

				// Instantiate a new CRF
				SequenceLabeller te = new SequenceLabeller();
				te.setEvalPeriod(100);
				te.setNFolds(0); // no cross-validation
				te.setTrainingFraction(1.0);
				te.setRandomSeed(0);
				cluster.setCrf(te);

				// Set training parameters
				HashMap<String, String> trainingInstances = trainingData[i][0];
				HashMap<String, Double> trainingPercentages = trainingData[i][1];
				//int effectiveTrainingIterations = trainingIterations;
				int effectiveTrainingIterations = (int) Math.ceil(((double) trainingIterations * (double) maxTrainingDataSize) / (double) trainingDataSizes[i]);
				System.out.println("[CLUSTERING] 	Training cluster " + cluster.getClusterId() + " for " + effectiveTrainingIterations + " iterations");
				// cluster.getCrf().setIterations(50); //fixed number of iterations
				cluster.getCrf().setIterations(effectiveTrainingIterations); // variable number of iterations
				cluster.getCrf().setGaussianVariance(20.0); // fixed variance
				
				// Train CRF model
				try {
					cluster.getCrf().trainModel(trainingInstances, numThreads, trainingPercentages);
					cluster.getCrf().saveModel("trainedModels/cluster" + cluster.getClusterId() + ".model.gz");
				} catch (IOException e) {
					e.printStackTrace();
				} catch (Exception e) {
					e.printStackTrace();
				}

				// clear previous assignments to this cluster
				cluster.clearInstances();
			}

			// Assign instances to closest cluster according to the log-likelihood
			System.out.println("[CLUSTERING] Re-assigning instances...");
			int j = 0;
			for (String id : this.dataFiles.keySet()) {
				j++;
				String instance = this.dataFiles.get(id);

				// Calculate log-likelihoods
				double[] loglikelihoods = new double[this.numClusters];
				for (int i = 0; i < this.numClusters; i++) {
					try {
						double ll = this.clusters[i].getCrf().loglikelihood(instance);
						System.out.println("[CLUSTERING] 	id=" + id + " cluster=" + this.clusters[i].getClusterId() + " ll=" + ll);
						loglikelihoods[i] = ll;
					} catch (IOException e) {
						e.printStackTrace();
						System.exit(-1);
					}
				}

				// Re-assign instance
				// Determine min and max for the normalization 
				double maxLogLikelihood = Integer.MIN_VALUE;
				double minLogLikelihood = Integer.MAX_VALUE;
				for (int i = 0; i < this.numClusters; i++) {
					if (loglikelihoods[i] > maxLogLikelihood) {
						maxLogLikelihood = loglikelihoods[i];
					}
					if (loglikelihoods[i] < minLogLikelihood) {
						minLogLikelihood = loglikelihoods[i];
					}
				}
				this.assignments.put(id, new ArrayList<Assignment>());
				double sumNormalizedLikelihoods = 0;
				for (int i = 0; i < this.numClusters; i++) {
					double normalizedLikelihood = (loglikelihoods[i] - minLogLikelihood) / (maxLogLikelihood - minLogLikelihood);
					sumNormalizedLikelihoods += normalizedLikelihood;
				}
				for (int i = 0; i < this.numClusters; i++) {
					double normalizedLikelihood = (loglikelihoods[i] - minLogLikelihood) / (maxLogLikelihood - minLogLikelihood);
					double proportion = normalizedLikelihood / sumNormalizedLikelihoods;
					currentGlobalLikelihoodFuzzy += (proportion*loglikelihoods[i]);
					System.out.println("[CLUSTERING] 	id=" + id + " cluster=" + this.clusters[i].getClusterId() 
							+ " normalizedLikelihood=" + normalizedLikelihood + " proportion=" + proportion);
					if (normalizedLikelihood >= 1.0) {
						this.assignments.get(id).add(new Assignment(id, clusters[i], 1.0));
					} else if (normalizedLikelihood >= 0.8) {
						this.assignments.get(id).add(new Assignment(id, clusters[i], normalizedLikelihood));
					} else {
						this.assignments.get(id).add(new Assignment(id, clusters[i], 0.0));
					}
					//this.assignments.get(id).add(new Assignment(id, clusters[i], normalizedLikelihood));
					//this.assignments.get(id).add(new Assignment(id, clusters[i], proportion));
				}
				Assignment bestAssignment = null;
				int bestAssignmentIndex = 0;
				int i=0;
				for (Assignment assignment : this.assignments.get(id)) {
					if (bestAssignment == null || assignment.getMembership() > bestAssignment.getMembership()) {
						bestAssignment = assignment;
						bestAssignmentIndex = i;
					}
					i++;
				}
				currentGlobalLikelihoodHard += loglikelihoods[bestAssignmentIndex];
				System.out.println("[CLUSTERING] 	Best assignment for instance " + id + " to cluster " + bestAssignment.getCluster().getClusterId()
						+ " (" + j + " of " + this.dataFiles.keySet().size() + ")");
				this.clusters[bestAssignment.getCluster().getClusterId()].addInstance(id, instance);

				// Check if the algorithm has not converged,
				// i.e. check if new cluster is the same as the one in the previous iteration
//				if (previousAssignments.get(id) == null || previousAssignments.get(id).getClusterId() != this.clusters[bestClusterFit].getClusterId()) {
//					converged = false;
//					System.out.println("Different cluster...");
//				} else {
//					System.out.println("Same as before...");
//				}
			}
			
			System.out.println("[CLUSTERING] 	iteration="+iteration+" previousGlobalLikelihoodHard=" + previousGlobalLikelihoodHard);
			System.out.println("[CLUSTERING] 	iteration="+iteration+" currentGlobalLikelihoodHard=" + currentGlobalLikelihoodHard);
			System.out.println("[CLUSTERING] 	iteration="+iteration+" previousGlobalLikelihoodFuzzy=" + previousGlobalLikelihoodFuzzy);
			System.out.println("[CLUSTERING] 	iteration="+iteration+" currentGlobalLikelihoodFuzzy=" + currentGlobalLikelihoodFuzzy);
			previousGlobalLikelihoodFuzzy = currentGlobalLikelihoodFuzzy;
			previousGlobalLikelihoodHard = currentGlobalLikelihoodHard;
			previousGlobalLikelihoodsHard[iteration] = currentGlobalLikelihoodHard;
			if (iteration > 2 && 
					previousGlobalLikelihoodsHard[iteration-2] > previousGlobalLikelihoodsHard[iteration] && 
					previousGlobalLikelihoodsHard[iteration-2] > previousGlobalLikelihoodsHard[iteration-1]) {
				converged = true;
			}
			
			if (converged) {
				System.out.println("[CLUSTERING] Algorithm converged!");
			}

			System.out.println("[CLUSTERING] New cluster assignments: ");
			pwAssignments.println("assignments" + iteration + " = {}");
			for (String id : this.assignments.keySet()) {
				Assignment bestAssignment = null;
				for (Assignment assignment : this.assignments.get(id)) {
					if (bestAssignment == null || assignment.getMembership() > bestAssignment.getMembership()) {
						bestAssignment = assignment;
					}
				}
				pwAssignments.println("assignments"+iteration+"['" + id + "'] = "
						+ bestAssignment.getCluster().getClusterId());
				System.out.println("[CLUSTERING] 	user=" + id + " -> cluster=" + bestAssignment.getCluster().getClusterId() + " membership="+bestAssignment.getMembership());
			}
			pwAssignments.println("cluster_assignments.append(assignments" + iteration + ")");
			pwAssignments.flush();

			if (converged) {
				System.out.println("[CLUSTERING] Algorithm converged!");
			}
		}
		pwAssignments.close();
	}

	public static void saveModel(CRFFuzzyKMeansClustering clustering,
			String outputFile) {
		System.out.println("[CLUSTERING] Saving model to file: " + outputFile);
		FileOutputStream fOut;
		try {
			fOut = new FileOutputStream(outputFile);
			ObjectOutputStream objOut = new ObjectOutputStream(fOut);
			objOut.writeObject(clustering);
			fOut.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public static CRFFuzzyKMeansClustering loadModel(String inputFile) {
		System.out
				.println("[CLUSTERING] Loading model from file: " + inputFile);
		FileInputStream fIn;
		try {
			fIn = new FileInputStream(inputFile);
			ObjectInputStream objIn = new ObjectInputStream(fIn);
			Object obj = objIn.readObject();

			if (obj instanceof CRFFuzzyKMeansClustering) {
				CRFFuzzyKMeansClustering clustering = (CRFFuzzyKMeansClustering) obj;
				return clustering;
			} else {
				throw new RuntimeException(
						"Input file is not a valid CRFKMeansClustering object: "
								+ inputFile);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	public static void main(String[] args) {
		int numClusters = 4;
		int maxIterations = 20;
		int numThreads = 4;
		CRFFuzzyKMeansClustering clustering = new CRFFuzzyKMeansClustering(numClusters, maxIterations, "data/synthetic_data_v2/stddev0.2", numThreads);
		clustering.clusterData();
		CRFFuzzyKMeansClustering.saveModel(clustering, "kcluster_synthetic_stddev0.2.model");

		clustering = CRFFuzzyKMeansClustering.loadModel("kcluster_synthetic_stddev0.2.model");
	}
	
	private class Assignment {
		private String id;
		private KCluster cluster;
		private double membership;
	
		public Assignment(String id, KCluster cluster, double membership) {
			super();
			this.id = id;
			this.cluster = cluster;
			this.membership = membership;
		}
		
		public String getId() {
			return id;
		}
		public void setId(String id) {
			this.id = id;
		}
		public KCluster getCluster() {
			return cluster;
		}
		public void setCluster(KCluster cluster) {
			this.cluster = cluster;
		}
		public double getMembership() {
			return membership;
		}
		public void setMembership(double membership) {
			this.membership = membership;
		}

		@Override
		public String toString() {
			return "Assignment [id=" + id + ", cluster=" + cluster
					+ ", membership=" + membership + "]";
		}
	}

}
