package pt.uc.dei.crf_ma.clustering;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Random;

import cc.mallet.types.MatrixOps;

import pt.uc.dei.crf_ma.SequenceLabeller;

public class CRFKMeansClustering {

	private static final long serialVersionUID = 8526472295622776148L;

	private Random randomGenerator;
	private int numClusters;
	private int maxIterations;
	private HashMap<String, String> dataFiles;
	private HashMap<String, Integer> dataSizes;
	private int numInstances;
	private int numThreads = 4;
	private KCluster[] clusters;

	public CRFKMeansClustering(int numClusters, int maxIterations,
			String inputFolder, int numThreads) {
		this.numClusters = numClusters;
		this.maxIterations = maxIterations;
		this.dataFiles = getDataFiles(inputFolder);
		this.numThreads = numThreads;
		this.numInstances = dataFiles.size();
		this.dataSizes = getDataSizes(this.dataFiles);
	}

	private HashMap<String, String> getDataFiles(String inputFolder) {
		HashMap<String, String> dataFiles = new HashMap<String, String>();

		File dir = new File(inputFolder);
		if (!inputFolder.endsWith("/")) {
			inputFolder += "/";
		}
		String[] children = dir.list();
		if (children == null) {
			throw new RuntimeException("Data directory not found: "
					+ inputFolder);
		} else {
			for (int i = 0; i < children.length; i++) {
				// Get filename of file or directory
				String filename = children[i];
				if (!filename.equals(".DS_Store")
						&& !filename.equals("._.DS_Store")) {
					dataFiles.put(filename, new String(inputFolder + filename));
				}
			}
		}

		return dataFiles;
	}

	private HashMap<String, Integer> getDataSizes(
			HashMap<String, String> dataFiles) {
		System.out.println("[CLUSTERING] Determining dataset sizes...");
		SequenceLabeller te = new SequenceLabeller();
		try {
			this.dataSizes = te.getDatasetSizes(dataFiles);
		} catch (IOException e) {
			e.printStackTrace();
		}
		return this.dataSizes;
	}

	public KCluster[] randomInitializationFull() {
		this.randomGenerator = new Random(0);
		System.out.println("[CLUSTERING] Num instances: " + this.numInstances);
		System.out.println("[CLUSTERING] Num clusters: " + this.numClusters);
		System.out
				.println("[CLUSTERING] Max iterations: " + this.maxIterations);
		System.out
				.println("[CLUSTERING] Initializing clusters with random data (full initialization)...");
		this.clusters = new KCluster[this.numClusters];
		for (int i = 0; i < this.numClusters; i++) {
			this.clusters[i] = new KCluster();
			SequenceLabeller te = new SequenceLabeller();

			// set the special options to the TermExtractor
			te.setEvalPeriod(100);
			te.setNFolds(0); // no cross-validation
			te.setIterations(20);
			te.setTrainingFraction(1.0);
			te.setRandomSeed(0);
			// te.setDebugPipes(true);

			this.clusters[i].setCrf(te);
		}
		for (String id : this.dataFiles.keySet()) {
			String dataFile = this.dataFiles.get(id);
			int rand = this.randomGenerator.nextInt(this.numClusters);
			this.clusters[rand].addInstance(id, dataFile);
		}

		for (KCluster cluster : this.clusters) {
			System.out.print("[CLUSTERING] 	Initilizing cluster "
					+ cluster.getClusterId() + " with the instances: ");
			for (String id : cluster.getInstances().keySet()) {
				System.out.print(id + ", ");
			}
			System.out.println();
		}

		return this.clusters;
	}

	public KCluster[] randomInitializationPartial() {
		this.randomGenerator = new Random(0);
		System.out.println("[CLUSTERING] Num instances: " + this.numInstances);
		System.out.println("[CLUSTERING] Num clusters: " + this.numClusters);
		System.out
				.println("[CLUSTERING] Max iterations: " + this.maxIterations);
		System.out
				.println("[CLUSTERING] Initializing clusters with random data (partial initialization)...");
		this.clusters = new KCluster[this.numClusters];
		ArrayList<Integer> pickedNumbers = new ArrayList<Integer>();
		ArrayList<String> ids = new ArrayList<String>(this.dataFiles.keySet());
		for (int i = 0; i < this.numClusters; i++) {
			this.clusters[i] = new KCluster();

			int rand = this.randomGenerator.nextInt(this.numInstances);
			while (pickedNumbers.contains(rand)) {
				rand = this.randomGenerator.nextInt(this.numInstances);
			}
			String id = ids.get(rand);
			this.clusters[i].addInstance(id, this.dataFiles.get(id));

			SequenceLabeller te = new SequenceLabeller();
			te.setEvalPeriod(100);
			te.setNFolds(0); // no cross-validation
			te.setTrainingFraction(1.0);
			te.setRandomSeed(0);
			// te.setDebugPipes(true);
			this.clusters[i].setCrf(te);
		}

		for (KCluster cluster : this.clusters) {
			System.out.print("[CLUSTERING] 	Initilizing cluster "
					+ cluster.getClusterId() + " with the instances: ");
			for (String id : cluster.getInstances().keySet()) {
				System.out.print(id + ", ");
			}
			System.out.println();
		}

		return this.clusters;
	}

	public KCluster[] randomInitializationSmart() {
		this.randomGenerator = new Random(0);
		System.out.println("[CLUSTERING] Num instances: " + this.numInstances);
		System.out.println("[CLUSTERING] Num clusters: " + this.numClusters);
		System.out
				.println("[CLUSTERING] Max iterations: " + this.maxIterations);
		System.out
				.println("[CLUSTERING] Initializing clusters with random data (smart initialization)...");
		this.clusters = new KCluster[this.numClusters];
		ArrayList<Integer> pickedNumbers = new ArrayList<Integer>();
		ArrayList<String> ids = new ArrayList<String>(this.dataFiles.keySet());
		double[][] loglikelihoods = new double[this.numClusters][this.numInstances];
		double[][] normalizedLoglikelihoods = new double[this.numClusters][this.numInstances];

		// Randomly assign first instance
		// Initialize cluster info
		int nextDataPoint = this.randomGenerator.nextInt(this.numInstances);
		for (int k = 0; k < this.numClusters; k++) {
			this.clusters[k] = new KCluster();
			pickedNumbers.add(nextDataPoint);
			String instance = ids.get(nextDataPoint);
			System.out.println("[CLUSTERING] Using instance=" + instance
					+ " to initialize cluster=" + k);
			this.clusters[k]
					.addInstance(instance, this.dataFiles.get(instance));
			SequenceLabeller te = new SequenceLabeller();
			te.setEvalPeriod(100);
			te.setNFolds(0); // no cross-validation
			te.setTrainingFraction(1.0);
			te.setRandomSeed(0);
			// te.setDebugPipes(true);
			this.clusters[k].setCrf(te);

			if (k == this.numClusters - 1) {
				continue; // skip next steps for the last cluster
			}

			// Train model
			te.setIterations(Integer.MAX_VALUE); // variable number of
													// iterations
			te.setGaussianVariance(1.0); // fixed variance
			try {
				te.trainModel(this.clusters[k].getInstances(), numThreads);
			} catch (IOException e) {
				e.printStackTrace();
			} catch (Exception e) {
				e.printStackTrace();
			}
			this.clusters[k].setCrf(te);

			// Calculate loglikelihoods
			for (int i = 0; i < this.numInstances; i++) {
				String inst = ids.get(i);
				String dataFile = this.dataFiles.get(inst);
				try {
					double ll = this.clusters[k].getCrf().loglikelihood(
							dataFile);
					System.out.println("[CLUSTERING] 	instance=" + inst
							+ " cluster=" + this.clusters[k].getClusterId()
							+ " ll=" + ll);
					loglikelihoods[k][i] = ll;
				} catch (IOException e) {
					e.printStackTrace();
					System.exit(1);
				}
			}

			// // Normalize loglikelihoods to the [0,1] interval
			// double min = MatrixOps.min(loglikelihoods[k]);
			// double max = MatrixOps.max(loglikelihoods[k]);
			// for (int i=0; i<loglikelihoods[k].length; i++) {
			// normalizedLoglikelihoods[k][i] = (loglikelihoods[k][i] - max) /
			// (max - min);
			// System.out.println("[CLUSTERING] 	normalized=" +
			// normalizedLoglikelihoods[k][i]);
			// }

			// Determine furthest data point
			double minLoglikelihoodSum = Double.MAX_VALUE;
			int furthestInstance = 0;
			for (int i = 0; i < loglikelihoods[k].length; i++) {
				double sum = 0;
				for (int j = 0; j <= k; j++) {
					sum += loglikelihoods[j][i];
				}
				sum /= (k + 1);
				System.out.println("[CLUSTERING] 	instance=" + ids.get(i)
						+ " cluster=" + this.clusters[k].getClusterId()
						+ " sum=" + sum);
				if (sum < minLoglikelihoodSum && !pickedNumbers.contains(i)) {
					minLoglikelihoodSum = sum;
					furthestInstance = i;
				}
			}
			nextDataPoint = furthestInstance;
		}

		for (KCluster cluster : this.clusters) {
			System.out.print("[CLUSTERING] 	Initilizing cluster "
					+ cluster.getClusterId() + " with the instances: ");
			for (String id : cluster.getInstances().keySet()) {
				System.out.print(id + ", ");
			}
			System.out.println();
		}

		return this.clusters;
	}

	public void clusterData() {
		//this.clusters = randomInitializationPartial();
		this.clusters = randomInitializationFull();
		//this.clusters = randomInitializationSmart();

		PrintWriter pwAssignments = null;
		try {
			pwAssignments = new PrintWriter("assignments.txt");
			pwAssignments.println("cluster_assignments = []");
			pwAssignments.println("assignments0 = {}");
			for (KCluster cluster : this.clusters) {
				for (String id : cluster.getInstances().keySet()) {
					pwAssignments.println("assignments0['" + id + "'] = "
							+ cluster.getClusterId());
				}
			}
			pwAssignments.println("cluster_assignments.append(assignments0)");
			pwAssignments.flush();
		} catch (FileNotFoundException e1) {
			e1.printStackTrace();
		}

		System.out.println("[CLUSTERING] Clustering started...");
		boolean converged = false;
		for (int iteration = 1; iteration <= this.maxIterations && !converged; iteration++) {
			System.out.println("[CLUSTERING] Iteration: " + iteration);
			converged = true;

			// Get previous cluster assignments
			HashMap<String, KCluster> previousAssignments = new HashMap<String, KCluster>();
			for (KCluster cluster : clusters) {
				for (String id : cluster.getInstances().keySet()) {
					if (previousAssignments.containsKey(id)) {
						throw new RuntimeException("Instance " + id
								+ " is assingned to more than one cluter!");
					}
					previousAssignments.put(id, cluster);
				}
			}

			int maxDataSize = Integer.MIN_VALUE;
			for (KCluster cluster : this.clusters) {
				int dataSize = 0;
				for (String id : cluster.getInstances().keySet()) {
					dataSize += this.dataSizes.get(id);
				}
				if (dataSize > maxDataSize) {
					maxDataSize = dataSize;
				}
			}

			// Train clusters
			//int trainingIterations = 10;
			int trainingIterations = (int) Math.pow(2, iteration-1);
			System.out.println("[CLUSTERING] 	Trying with training iterations=" + trainingIterations);
			for (KCluster cluster : this.clusters) {
				int dataSize = 0;
				for (String id : cluster.getInstances().keySet()) {
					dataSize += this.dataSizes.get(id);
				}

				// Instantiate a new CRF
					SequenceLabeller te = new SequenceLabeller();
					te.setEvalPeriod(100);
					te.setNFolds(0); // no cross-validation
					te.setTrainingFraction(1.0);
					te.setRandomSeed(0);
					cluster.setCrf(te);

				// Train model
				//int effectiveTrainingIterations = trainingIterations;
				int effectiveTrainingIterations = (int) Math.ceil(((double) trainingIterations * (double) maxDataSize) / (double) dataSize);
				System.out.println("[CLUSTERING] 	Training cluster " + cluster.getClusterId() + " for " + effectiveTrainingIterations + " iterations");
				// cluster.getCrf().setIterations(50); //fixed number of iterations
				cluster.getCrf().setIterations(effectiveTrainingIterations); // variable number of iterations
				cluster.getCrf().setGaussianVariance(20.0); // fixed variance
				try {
					cluster.getCrf().trainModel(cluster.getInstances(), numThreads);
					cluster.getCrf().saveModel("trainedModels/cluster" + cluster.getClusterId() + ".model.gz");
				} catch (IOException e) {
					e.printStackTrace();
				} catch (Exception e) {
					e.printStackTrace();
				}

				// clear previous assignments to this cluster
				cluster.clearInstances();
			}

			// Assign instances to closest cluster according to the
			// log-likelihood
			System.out.println("[CLUSTERING] Re-assigning instances...");
			int j = 0;
			for (String id : this.dataFiles.keySet()) {
				j++;
				String instance = this.dataFiles.get(id);

				// Calculate log-likelihoods
				double[] loglikelihoods = new double[this.numClusters];
				for (int i = 0; i < this.numClusters; i++) {
					try {
						double ll = this.clusters[i].getCrf().loglikelihood(instance);
						System.out.println("[CLUSTERING] 	id=" + id + " cluster=" + this.clusters[i].getClusterId() + " ll=" + ll);
						loglikelihoods[i] = ll;
					} catch (IOException e) {
						e.printStackTrace();
						System.exit(-1);
					}
				}

				// Re-assign instance
				double maxLogLikelihood = Integer.MIN_VALUE;
				int bestClusterFit = -1;
				for (int i = 0; i < this.numClusters; i++) {
					if (loglikelihoods[i] > maxLogLikelihood) {
						maxLogLikelihood = loglikelihoods[i];
						bestClusterFit = i;
					}
				}
				System.out.print("[CLUSTERING] 	Assigning instance " + id + " to cluster " + this.clusters[bestClusterFit].getClusterId()
						+ " (" + j + " of " + this.dataFiles.keySet().size() + ") - ");
				this.clusters[bestClusterFit].addInstance(id, instance);

				// Check if the algorithm has not converged,
				// i.e. check if new cluster is the same as the one in the
				// previous iteration
				if (previousAssignments.get(id) == null || previousAssignments.get(id).getClusterId() != this.clusters[bestClusterFit].getClusterId()) {
					converged = false;
					System.out.println("Different cluster...");
				} else {
					System.out.println("Same as before...");
				}
			}
			
			if (converged) {
				System.out.println("[CLUSTERING] Algorithm converged!");
			}

			System.out.println("[CLUSTERING] New cluster assignments: ");
			pwAssignments.println("assignments" + iteration + " = {}");
			for (KCluster cluster : this.clusters) {
				String data = "";
				for (String id : cluster.getInstances().keySet()) {
					data += id + ", ";
					pwAssignments.println("assignments" + iteration + "['" + id + "'] = " + cluster.getClusterId());
				}
				System.out.println("[CLUSTERING] 	Cluster " + cluster.getClusterId() + ": " + data);
			}
			pwAssignments.println("cluster_assignments.append(assignments" + iteration + ")");
			pwAssignments.flush();
		}
		pwAssignments.close();
	}

	public static void saveModel(CRFKMeansClustering clustering,
			String outputFile) {
		System.out.println("[CLUSTERING] Saving model to file: " + outputFile);
		FileOutputStream fOut;
		try {
			fOut = new FileOutputStream(outputFile);
			ObjectOutputStream objOut = new ObjectOutputStream(fOut);
			objOut.writeObject(clustering);
			fOut.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public static CRFKMeansClustering loadModel(String inputFile) {
		System.out
				.println("[CLUSTERING] Loading model from file: " + inputFile);
		FileInputStream fIn;
		try {
			fIn = new FileInputStream(inputFile);
			ObjectInputStream objIn = new ObjectInputStream(fIn);
			Object obj = objIn.readObject();

			if (obj instanceof CRFKMeansClustering) {
				CRFKMeansClustering clustering = (CRFKMeansClustering) obj;
				return clustering;
			} else {
				throw new RuntimeException(
						"Input file is not a valid CRFKMeansClustering object: "
								+ inputFile);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	public static void main(String[] args) {
		int numClusters = 4;
		int maxIterations = 20;
		int numThreads = 4;
		CRFKMeansClustering clustering = new CRFKMeansClustering(numClusters, maxIterations, "data/synthetic_data_v2/stddev0.2", numThreads);
		clustering.clusterData();
		CRFKMeansClustering.saveModel(clustering, "kcluster_synthetic_stddev0.2.model");

		clustering = CRFKMeansClustering.loadModel("kcluster_synthetic_stddev0.2.model");
	}

}
