package pt.uc.dei.crf_ma.clustering;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;

import pt.uc.dei.crf_ma.SequenceLabeller;

public class CRFProgressiveKMeansClustering {
	
	private static final long serialVersionUID = 8526472295622776149L;

	private int numClusters;
	private int maxIterations;
	private HashMap<String, String> dataFiles;
	private HashMap<String, Integer> dataSizes;
	private int numInstances;
	private int numThreads = 4;
	private KCluster[] clusters;
	
	public CRFProgressiveKMeansClustering(int numClusters, int maxIterations, String inputFolder, int numThreads) {
		this.numClusters = numClusters;
		this.maxIterations = maxIterations;
		this.dataFiles = getDataFiles(inputFolder);
		this.numThreads = numThreads;
		this.numInstances = dataFiles.size();
		this.dataSizes = getDataSizes(this.dataFiles);
	}
	
	private HashMap<String, String> getDataFiles(String inputFolder) {
		HashMap<String, String> dataFiles = new HashMap<String, String>();
		
		File dir = new File(inputFolder);
		if (!inputFolder.endsWith("/")) {
			inputFolder += "/";
		}
		String[] children = dir.list();
		if (children == null) {
		    throw new RuntimeException("Data directory not found: " + inputFolder);
		} else {
		    for (int i=0; i<children.length; i++) {
		        // Get filename of file or directory
		        String filename = children[i];
		        if (!filename.equals(".DS_Store") && !filename.equals("._.DS_Store")) {
		        	dataFiles.put(filename, new String(inputFolder + filename));
		        }
		    }
		}
		
		return dataFiles;
	}
	
	private HashMap<String, Integer> getDataSizes(HashMap<String, String> dataFiles) {
		System.out.println("[CLUSTERING] Determining dataset sizes...");
		SequenceLabeller te = new SequenceLabeller();
		try {
			this.dataSizes = te.getDatasetSizes(dataFiles);
		} catch (IOException e) {
			e.printStackTrace();
		}
		return this.dataSizes;
	}
	
	public KCluster[] initializeClusters() {
		System.out.println("[CLUSTERING] Num instances: " + this.numInstances);
		System.out.println("[CLUSTERING] Num clusters: " + this.numClusters);
		System.out.println("[CLUSTERING] Max iterations: " + this.maxIterations);
		System.out.println("[CLUSTERING] Initializing clusters with no data...");
		this.clusters = new KCluster[this.numClusters];
		for (int i=0; i<this.numClusters; i++) {
			this.clusters[i] = new KCluster();
			SequenceLabeller te = new SequenceLabeller();
			
			// set the special options to the TermExtractor
			te.setEvalPeriod(100);
			te.setNFolds(0); // no cross-validation
			te.setIterations(20);
			te.setTrainingFraction(1.0);
			te.setRandomSeed(0);
			te.setGaussianVariance(50.0);
			//te.setDebugPipes(true);
			
			this.clusters[i].setCrf(te);
		}
		
		return this.clusters;
	}
	
	public void clusterData() {
		this.clusters = initializeClusters();
		
		PrintWriter pwAssignments = null;
		try {
			pwAssignments = new PrintWriter("assignments.txt");
			pwAssignments.println("cluster_assignments = []");
			pwAssignments.println("assignments0 = {}");
			for (KCluster cluster : this.clusters) {
				for (String id : cluster.getInstances().keySet()) {
					pwAssignments.println("assignments0['" + id + "'] = " + cluster.getClusterId());
				}
			}
			pwAssignments.println("cluster_assignments.append(assignments0)");
			pwAssignments.flush();
		} catch (FileNotFoundException e1) {
			e1.printStackTrace();
		}
		
		ArrayList<String> unassignedInstances = new ArrayList<String>();
		ArrayList<String> assignedInstances = new ArrayList<String>();
		for (String id : this.dataFiles.keySet()) {
			unassignedInstances.add(id);
		}
		
		// Initialize clusters
		// Add one instance to each cluster (the instance with more data)
		for (int i=0; i<this.numClusters && unassignedInstances.size() > 0; i++) {
			// Get largest data instance, i.e. the instance with more data
			String largestInstance = null;
			int largestInstanceSize = 0;
			for (String id : unassignedInstances) {
				int size = this.dataSizes.get(id);
				if (size > largestInstanceSize) {
					largestInstance = id;
					largestInstanceSize = size;
				}
			}
			unassignedInstances.remove(largestInstance);
			this.clusters[i].addInstance(largestInstance, this.dataFiles.get(largestInstance));
			assignedInstances.add(largestInstance);
		}
		
		System.out.println("[CLUSTERING] Clustering started...");
		boolean converged = false;
		for (int iteration = 1; iteration <= this.maxIterations && !converged && unassignedInstances.size() > 0; iteration++) {
			System.out.println("[CLUSTERING] Iteration: " + iteration);
			converged = false;
			
			// Get previous cluster assignments
			HashMap<String, KCluster> previousAssignments = new HashMap<String, KCluster>();
			for (KCluster cluster : clusters) {
				for (String id : cluster.getInstances().keySet()) {
					if (previousAssignments.containsKey(id)) {
						throw new RuntimeException("Instance " + id + " is assingned to more than one cluter!");
					}
					previousAssignments.put(id, cluster);
				}
			}

			int maxDataSize = Integer.MIN_VALUE;
			for (KCluster cluster : this.clusters) {
				int dataSize = 0;
				for (String id : cluster.getInstances().keySet()) {
					dataSize += this.dataSizes.get(id);
				}
				if (dataSize > maxDataSize) {
					maxDataSize = dataSize;
				}
			}
			
			// Train clusters
			int trainingIterations = (int) Math.pow(2, iteration);
			//int trainingIterations = 10;
			for (KCluster cluster : this.clusters) {
				int dataSize = 0;
				for (String id : cluster.getInstances().keySet()) {
					dataSize += this.dataSizes.get(id);
				}
				int effectiveTrainingIterations = (trainingIterations * maxDataSize) / dataSize; 
				
				// Train model
				System.out.println("[CLUSTERING] 	Training cluster " + cluster.getClusterId() + " for " + effectiveTrainingIterations + " iterations");
				//cluster.getCrf().setIterations(10); //fixed number of iterations
				cluster.getCrf().setIterations(effectiveTrainingIterations); // variable number of iterations
				cluster.getCrf().setGaussianVariance(50.0); //fixed variance
				try {
					cluster.getCrf().trainModel(cluster.getInstances(), numThreads);
					cluster.getCrf().saveModel("trainedModels/cluster" + cluster.getClusterId() + ".model.gz");
				} catch (IOException e) {
					e.printStackTrace();
				} catch (Exception e) {
					e.printStackTrace();
				}
				
				// clear previous assignments to this cluster
				cluster.clearInstances(); 
			}
			
			// Consider new data in the clustering: choose the largest instances
			for (int i=0; i<this.numClusters && unassignedInstances.size() > 0; i++) {
				// Get largest data instance, i.e. the instance with more data
				String largestInstance = null;
				int largestInstanceSize = 0;
				for (String id : unassignedInstances) {
					int size = this.dataSizes.get(id);
					if (size > largestInstanceSize) {
						largestInstance = id;
						largestInstanceSize = size;
					}
				}
				if (largestInstanceSize < 500) {
					unassignedInstances.clear();
				}
				unassignedInstances.remove(largestInstance);
				assignedInstances.add(largestInstance);
			}
			
			// Assign instances to closest cluster according to the log-likelihood
			System.out.println("[CLUSTERING] Re-assigning instances...");
			int j = 0;
			for (String id : assignedInstances) {
				j++;
				String instance = this.dataFiles.get(id);
				
				// Calculate log-likelihoods
				double[] loglikelihoods = new double[this.numClusters];
				for (int i= 0; i<this.numClusters; i++) {
					try {
						double ll = this.clusters[i].getCrf().loglikelihood(instance);
						System.out.println("[CLUSTERING] 	id=" + id + " cluster=" + this.clusters[i].getClusterId() + " ll=" + ll);
						loglikelihoods[i] = ll;
					} catch (IOException e) {
						e.printStackTrace();
						System.exit(-1);
					}
				}
				
				// Re-assign instance
				double maxLogLikelihood = Integer.MIN_VALUE;
				int bestClusterFit = -1;
				for (int i = 0; i<this.numClusters; i++) {
					if (loglikelihoods[i] > maxLogLikelihood) {
						maxLogLikelihood = loglikelihoods[i];
						bestClusterFit = i;
					}
				}
				System.out.println("[CLUSTERING] 	Assigning instance " + id + 
						" to cluster " + this.clusters[bestClusterFit].getClusterId() + 
						" (" + j + " of " + this.dataFiles.keySet().size() + ")");
				this.clusters[bestClusterFit].addInstance(id, instance);
				
				// Check if the algorithm has not converged, 
				// i.e. check if new cluster is the same as the one in the previous iteration
				if (previousAssignments.get(id) == null || previousAssignments.get(id).getClusterId() != this.clusters[bestClusterFit].getClusterId()) {
					converged = false;
				}
			}
			if (converged) {
				System.out.println("[CLUSTERING] Algorithm converged!");
			}
			
			System.out.println("[CLUSTERING] New cluster assignments: ");
			pwAssignments.println("assignments"+iteration+" = {}");
			for (KCluster cluster : this.clusters) {
				String data = "";
				for (String id : cluster.getInstances().keySet()) {
					data += id + ", ";
					pwAssignments.println("assignments"+iteration+"['" + id + "'] = " + cluster.getClusterId());
				}
				System.out.println("[CLUSTERING] 	Cluster " + cluster.getClusterId() + ": " + data);
			}
			pwAssignments.println("cluster_assignments.append(assignments"+iteration+")");
			pwAssignments.flush();
		}
		pwAssignments.close();
	}
	
	public static void saveModel(CRFProgressiveKMeansClustering clustering, String outputFile) {
		System.out.println("[CLUSTERING] Saving model to file: " + outputFile);
		FileOutputStream fOut;
		try {
			fOut = new FileOutputStream(outputFile);
			ObjectOutputStream objOut = new ObjectOutputStream (fOut);
			objOut.writeObject(clustering);
			fOut.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public static CRFProgressiveKMeansClustering loadModel(String inputFile) {
		System.out.println("[CLUSTERING] Loading model from file: " + inputFile);
		FileInputStream fIn;
		try {
			fIn = new FileInputStream(inputFile);
			ObjectInputStream objIn = new ObjectInputStream(fIn);
			Object obj = objIn.readObject();

			if (obj instanceof CRFProgressiveKMeansClustering) {
				CRFProgressiveKMeansClustering clustering = (CRFProgressiveKMeansClustering) obj;
				return clustering;
			} else {
				throw new RuntimeException("Input file is not a valid CRFKMeansClustering object: " + inputFile);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	public static void main(String[] args) {
		int numClusters = 4;
		int maxIterations = 20;
		int numThreads = 4;
		CRFProgressiveKMeansClustering clustering = new CRFProgressiveKMeansClustering(numClusters, maxIterations, "data/synthetic_data_stddev0.2", numThreads);
		clustering.clusterData();
		CRFProgressiveKMeansClustering.saveModel(clustering, "kcluster.model");
		
		clustering = CRFProgressiveKMeansClustering.loadModel("kcluster.model");
	}

}
