package pt.uc.dei.crf_ma.evaluator;

import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import pt.uc.dei.crf_ma.SequenceLabeller;
import pt.uc.dei.crf_ma.types.TermExtractorInstance;
import cc.mallet.fst.Transducer;
import cc.mallet.fst.TransducerEvaluator;
import cc.mallet.fst.TransducerTrainer;
import cc.mallet.types.FeatureVector;
import cc.mallet.types.InstanceList;
import cc.mallet.types.Sequence;


/**
 * Evaluates a transducer model, computes the precision, recall and F1 scores;
 * considers segments that span across multiple tokens.
 */
public class WordEvaluator extends TransducerEvaluator
{
	private Logger logger = LoggerFactory.getLogger(WordEvaluator.class);

	// equals() is called on these objects to determine if this token is the start or continuation of a segment.
	// A tag not equal to any of these is an "other".
	// is not part of the segment).
	Object[] segmentStartTags;
	Object[] segmentContinueTags;
	Object[] segmentStartOrContinueTags;
	
	public WordEvaluator (InstanceList[] instanceLists, String[] instanceListDescriptions,
			Object[] segmentStartTags, Object[] segmentContinueTags)
	{
		super (instanceLists, instanceListDescriptions);
		this.segmentStartTags = segmentStartTags;
		this.segmentContinueTags = segmentContinueTags;
		assert (segmentStartTags.length == segmentContinueTags.length);
	}
	
	public WordEvaluator (InstanceList instanceList1, String description1,
			Object[] segmentStartTags, Object[] segmentContinueTags)
	{
		this (new InstanceList[] {instanceList1}, new String[] {description1},
				segmentStartTags, segmentContinueTags);
	}
	
	public WordEvaluator (InstanceList instanceList1, String description1,
			InstanceList instanceList2, String description2,
			Object[] segmentStartTags, Object[] segmentContinueTags)
	{
		this (new InstanceList[] {instanceList1, instanceList2}, new String[] {description1, description2},
				segmentStartTags, segmentContinueTags);
	}

	public WordEvaluator (InstanceList instanceList1, String description1,
			InstanceList instanceList2, String description2,
			InstanceList instanceList3, String description3,
			Object[] segmentStartTags, Object[] segmentContinueTags)
	{
		this (new InstanceList[] {instanceList1, instanceList2, instanceList3}, new String[] {description1, description2, description3},
				segmentStartTags, segmentContinueTags);
	}

	public void evaluateInstanceList(TransducerTrainer tt, InstanceList data,
			String description) {
		evaluateInstanceListWithResults(tt, data, description, false);
	}

	public EvaluatorResults evaluateInstanceListWithResults (TransducerTrainer tt, InstanceList data, String description, boolean silent)
	{
	  	Transducer model = tt.getTransducer();
	    int numCorrectTokens, totalTokens;
	    int numTrueSegments = 0;
	    int numPredictedSegments = 0; 
	    int numCorrectSegments = 0;
	    int allIndex = segmentStartTags.length;
	    
	    Map<String, Map<String, Object>> articleKeywordsTrue = new HashMap<String, Map<String, Object>>();
	    Map<String, Map<String, Object>> articleKeywordsPred = new HashMap<String, Map<String, Object>>();
	    
	    for (int i=0;i<data.size();i++) {
	    	TermExtractorInstance instance = (TermExtractorInstance) data.get(i);
	        Sequence input = (Sequence) instance.getData();
	        Sequence trueOutput = (Sequence) instance.getTarget();
	        Sequence predOutput = model.transduce (input);
	        
	        String instanceId = instance.getInstanceId();
	        
	    	// get all the true keywords
	        Map<String, Object> kw1 = getKeywords(input, trueOutput);
	    	Map<String, Object> kw2 = getKeywords(input, predOutput);
	    	
	        addKeywords2Article(articleKeywordsTrue, instanceId, kw1);
	        addKeywords2Article(articleKeywordsPred, instanceId, kw2);
	    }
	    
	    assert(articleKeywordsTrue.size() == articleKeywordsPred.size());
	    for (String key : articleKeywordsTrue.keySet()) {
	    	Map<String, Object> kw1 = articleKeywordsTrue.get(key);
	    	Map<String, Object> kw2 = articleKeywordsPred.get(key);
	    	
	    	// global statistics
	    	numPredictedSegments += kw2.size();
	    	numTrueSegments += kw1.size();
	    	
	    	// check how many of them are correct
	    	java.util.Iterator<String> it = kw2.keySet().iterator();
	    	while (it.hasNext())
	    		if (kw1.containsKey(it.next()))
	    			numCorrectSegments++;
	    	
	    }
	    
	    DecimalFormat f = new DecimalFormat ("0.####");
		double precision = numPredictedSegments == 0 ? 1 : ((double)numCorrectSegments) / numPredictedSegments;
		double recall = numTrueSegments == 0 ? 1 : ((double)numCorrectSegments) / numTrueSegments;
		double f1 = recall+precision == 0.0 ? 0.0 : (2.0 * recall * precision) / (recall + precision);
		if (!silent) 
			logger.info (" "+description+" precision="+f.format(precision)+" recall="+f.format(recall)+" f1="+f.format(f1));
		
		return new EvaluatorResults(0,
				0, numCorrectSegments,
				numPredictedSegments, numTrueSegments, Double.NaN,
				precision, recall, f1);
	}
	
	private void addKeywords2Article(Map<String, Map<String, Object>> articleKeywords, String instanceId, Map<String, Object> kw) {
		Map<String, Object> result = articleKeywords.get(instanceId);
		//System.out.println("adding keywords to article id ="+articleId.toString());
		
		if (result == null)
			result = new HashMap<String, Object>();
		
		for (String key: kw.keySet())
			result.put(key, null);
		
		articleKeywords.put(instanceId, result);
	}
	
	private void printKeywords(Map<String, Object> kw) {
		Set<String> keywords = kw.keySet();
		
		java.util.Iterator<String> it = keywords.iterator();
		while (it.hasNext()) 
			System.out.print(it.next()+" | ");
		
		System.out.println("");
	}
	
	private Object isStartTag(Object tag) {
		for (int j=0;j<segmentStartTags.length;j++) {
			if (tag.equals(segmentStartTags[j])) {
				return segmentStartTags[j];
			}
		}
		
		return null;
	}
	
	private Object isContinueTag(Object tag) {
		for (int j=0;j<segmentContinueTags.length;j++) {
			if (tag.equals(segmentContinueTags[j])) {
				return segmentContinueTags[j];
			}
		}
		
		return null;
	}
	
	/* we should try to get the word in another way.. 
	 * this is a very bad approach
	 */
	private Map<String, Object> getKeywords(Sequence input, Sequence target) {
		Map<String, Object> result = new HashMap<String, Object>();
		
		// boolean flag to tell us if we are 
		// "inside" a keywork window
		boolean inside = false;
		String keyword = new String();
		Object startTag = null;
		Object continueTag=null;
		
		for (int i=0;i<input.size();i++) {
			
			continueTag = isContinueTag(target.get(i));	
			startTag = isStartTag(target.get(i));

			if (continueTag!=null || startTag!=null)
				result.put(getWord((FeatureVector) input.get(i)), null);
			
		}
		
		return result;
	}
		
	private String getWord(FeatureVector fv) {
		String[] features = fv.toString().split("\n");
      	String originalToken = null;
      	
      	for (String feature : features) {
      		if(Pattern.matches(SequenceLabeller.TOKEN_FNAME+".*", feature) && !feature.contains("@")) {
      			originalToken = feature.replace(SequenceLabeller.TOKEN_FNAME, "").replaceAll("\\([0-9]+\\)=[0-9]+\\.[0-9]+", "");
      			return originalToken;
      		}
      	}
      	
      	return null;
	}

}
