
package pt.uc.dei.crf_ma.pipe;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.HashMap;
import java.util.zip.GZIPInputStream;

import cc.mallet.pipe.Pipe;
import cc.mallet.types.Instance;
import cc.mallet.types.Token;
import cc.mallet.types.TokenSequence;

/**
 * This pipe is used to calc and add an feature
 * that represents the tfidf of the Tokens.
 * The Corpus used for the IDF is calculated using the gensim module 
 * executed over the Wikipedia datadump.
 * If the token is too rare or to common it should not appear in the corpus.
 *
 * @author João Oliveirinha
 */
public class TFIDF extends Pipe implements Serializable {

    
	private static final long serialVersionUID = 1L;
	private static final String IDFPATH = "/pipes/idf.gz";
	
	private String featureName;
	private HashMap<String, Double> idfmap; 
	
	public TFIDF(String featureName) {
		this.featureName = featureName;
		
		try {
			idfmap = loadIDFMapping(IDFPATH);
		} catch (IOException e) {
			e.printStackTrace();
			throw new RuntimeException(e.getMessage());
		}
	}
	
	private HashMap<String, Double> loadIDFMapping(String IDFPATH) throws IOException {
		HashMap<String, Double> result = new HashMap<String, Double>();
		
		InputStream modelFile = new GZIPInputStream(getClass().getResourceAsStream(IDFPATH));
		BufferedReader reader = new BufferedReader(new InputStreamReader(modelFile));
		
		String line = null;
		while ((line = reader.readLine())!=null) {
			String [] fields = line.split("\t");
			String word = fields[0];
			double idf = Double.parseDouble(fields[1]);
			
			result.put(word.toLowerCase(), idf);
		}
		
		return result;
 	}
	
    @Override
    public Instance pipe(Instance carrier) {
        TokenSequence ts = (TokenSequence) carrier.getData();
        
        HashMap<String, Integer> termFrequencies = new HashMap<String, Integer>();
        for (int i=0;i<ts.size();i++) {
        	String word = ts.get(i).getText().toLowerCase();
        	Integer c = termFrequencies.get(word);
        	
        	int conta = 0;
        	if (c!=null)
        		conta = c.intValue();
        	
        	termFrequencies.put(word, new Integer(conta+1));
        }
        
        for(int i = 0; i<ts.size(); i++) {
        	Token t = ts.get(i);
        	
        	Double idf = idfmap.get(t.getText().toLowerCase());
        	if (idf != null) {
        		double tf = termFrequencies.get(t.getText().toLowerCase()).intValue() / (double) ts.size();
        		double tfidf =  tf * idf.doubleValue();
        		t.setFeatureValue(featureName, tfidf);
        	} 
        }
      
        return carrier;
    }

	// Serialization 
	
	private void writeObject (ObjectOutputStream out) throws IOException {
		out.writeLong(serialVersionUID);
		out.writeObject(featureName);
	}
	
	private void readObject (ObjectInputStream in) throws IOException, ClassNotFoundException {
		long version = in.readLong ();
		// we should check here if this is the correct version
		
		featureName = (String) in.readObject();
		idfmap = loadIDFMapping(IDFPATH);
	}
}
